** For best results, view this file with: Courier New / Regular / 10
** TabWidth set to: 4
*******************************************************************************
** Program Name  : GenDBCXm.PRG (nee GENDBC.PRG and GenDBCX)
** Version       : 2.0  && 1.12
** Creation Dates: 94.12.01 (GENDBC.PRG)  97.02.01 (GenDBCX.PRG)  jme 29/05/2008 gendbcxm
**
** Purpose:
**   To take an existing FoxPro 9 database and generate an output
**	  program that can be used to "re-create" that database.
**   plus : generate 'access classes' to each table. An 'access class'
**   is a class that have methods like : OPEN (USE), CLOSE (USE IN),
**   ADD, MODIFY, DELETE, SEARCH ....
**
** Parameters:
**   tcOutFile   - The name of the output file.  This can contain path 
**                 information and an extension.
**                 If no extension is supplied, one will be provided.
**
**   tcClassName - The name of the class to create.  If not supplied,
**                 the default 'GenDBCX' will be used.
**
**   tlSeparateSPObjCode
**               - A flag that tells the program whether to append the
**                 Stored Procedures object code to the .krt file that
**                 holds the SP source code (the default), or write it
**                 to a separate .kro file.
**
** Modification History:
**            (KRT = Somebody at MS; SEA = Steve Arnott, SEA Drive;
**             SAS = Steve Sawyer)
**
** GENDBC:
** 1994.12.01 KRT  Created program, runs on Build 329 of VFP
** 1994.12.02 KRT  Added GetView function and cleaned up all code
** 1994.12.05 KRT  Modified some areas and verified run on Build 335
** 1994.12.06 KRT  Added things for international conversion
** 1994.12.08 KRT  Added function ADBOBJECTS() to code
** 1994.12.12 KRT  Added commands COPY PROCEDURES TO
** 1995.01.04 KRT  Added connection properties
** 1995.01.05 KRT  Added support for long filenames - thierryp
** 1995.01.05 KRT  Added support for RI
** 1995.01.06 KRT  Added support for MS-DOS short filenames (NAME clause)
** 1995.01.08 KRT  Added status bar line
** 1995.01.26 KRT  Fixed a few file bugs and localization bug
** 1995.02.19 KRT  Took advantage of AFIELDS() command
** 1995.02.22 KRT  Fixed AUSED() bug
** 1995.03.01 KRT  Removed ON ERROR problem
** 1995.03.20 KRT  Fixed "Exclusive" Error / Procedures created before
**                 Tables
** 1995.03.22 KRT  Allowed user to open database if one is not set current
**                 / Set SAFETY OFF in resulting code to prevent errors
**                   when validating rules/triggers that don't exist
** 1995.04.07 KRT  Put any database procedures into a seperate file to
**                 prevent "Program Too Large" errors
** 1995.04.20 KRT  Only change SAFETY when appending procedures
**
** 3.0b - Changes/Fixes
** 1995.09.15 KRT  Changed ADBOBJECTS() to DBF() in GETTABLE procedure.
**                 This allows for "real" table names instead of alias'
** 1995.09.15 KRT  Take into account CR+LF in comments for fields
** 1995.09.15 KRT  Store Source and Object code into external file
**                 So the code can be executed from the run-time version
** 1995.10.23 KRT  Changed DBF(cTableName ) to DBF(ALIAS() ) because
**                 VFP will automatically add underscores where spaces
**                 should be and I need to know what it did to the alias
**                 [Regress]
** 1995.10.25 KRT  Added OVERWRITE to append memo's [Regress]
** 1995.10.25 KRT  Added support for CR+LF in comments for Table
** 1995.10.25 KRT  Close all tables in generated code before adding
**                 Source and Object code from external file[Regress]
** 1995.10.25 KRT  Added warning about filter indices on Unique and
**                 Candidate keys (Not supported via language )
** 1995.12.20 KRT  Better lookup for adding RI information to database
** 1995.12.20 KRT  Added support for filter expressions on CANDIDATE keys
** 
** 5.0 - Changes/Fixes
** 1996.03.27 KRT  Added "exact match" support for Locate command
** 1996.04.12 KRT  Added new properties for Views, Fields and Connections
** 1996.05.14 KRT  Adjusted for some logical properties return spaces
** 1996.05.16 KRT  Added even more new properties for Views
** 1996.05.16 KRT  Add M. in front of all memory variables to prevent
**                 confusion with table fields and names
** 1996.06.01 KRT  Added support for Collate sequence on index files
** 1996.06.26 KRT  Added support for ParameterList in Views
** 1996.07.19 KRT  Added support for Offline Views
** 1996.08.07 KRT  Added support for comments and default values in views
**
** GenDBCX:
** 1997.02.01 SEA  Changed "Get Relations" code to build output in a
**                 memory variable to avoid creating 100+MB memo files
**                 when working with large databases
** 1997.02.24 SEA  Added code to avoid the dreaded "Program too large"
**                 error message by making (almost) everything in the
**                 database re-creation program a procedure
** 1997.02.24 SEA  Replaced usage of "Program" memo field with writing
**                 directly to a temporary file to prevent "Out of disk
**                 space" error caused by memo file bloat
** 1997.02.25 SEA  Changed format of CREATE TABLE output code to improve 
**                 (I hope) its readability
** 1997.03.23 SEA  Added code to create the DBC regeneration program as a
**                 class, so that you can recreate specific tables, views
**                 and connections by calling the output class' methods.
**                 To recreate the entire database, you either run the 
**                 program, or call the GenerateAll method.
** 1997.03.24 SEA  Added code to create a class Init routine that checks
**                 for the existence of the target database container and,
**                 if it finds it, loads the names of its tables into an
**                 array that the table creation methods search.  If the
**                 table name exists in the dbc, it's deleted and recreated.
** 1997.04.04 SEA  Added code to allow user to pass the name of the method
**                 to call as a parameter, rather than forcing the user to
**                 instantiate a GenDBCX object.
** 1997.05.01 SAS  Added code to recreate the dbc in two steps, with a hook
**                 in the middle that allows the user to append existing data
**                 to the new database, or do anything else that the user might
**                 require upon creation of the new database.
** 1997.07.29 SEA  Fixed some minor formatting errors.
**                 Fixed the RI "error" message line so that the number of the
**                 line in the regeneration program is displayed, rather than
**                 the number of the line in this program.
** 1997.09.05 SEA  Fixed procedure GetView to handle 5.0-specific properties
**                 when running under VFP 3.0.  Thanks to Nancy Folsom for
**                 catching this!
** 1997.09.05 SEA  Added code to handle table, view and connection names that
**                 contain spaces.
** 1997.10.01 SEA  Added parameter tlSeparateSPObjCode to allow the user to
**                 specify that the Stored Procedures object code is to be
**                 output to a separate .kro file, rather than appended to
**                 the .krt file that holds the SP source code.  Apparently,
**                 the inclusion of object code in the .krt prevents the file
**                 from being compared to a previous version in Visual
**                 SourceSafe projects.
** 1997.10.15 SEA  Added code to allow the user to sort the arrays that hold
**                 the table, view, connection and relation names alphabetically,
**                 so that they appear in alphabetical order in the output code.
** 1997.11.01 SEA  Added ability to control the appearance of the output code
**                 through the use of constants that control code indentation
**                 and the separation of field names from their data types
**                 in CREATE TABLE commands.
** 1997.11.01 SEA  Added code in procedure GetTable to handle "Program line too
**                 long" error which occurs when a table has a large number
**                 of fields.  Thanks to Ken Lee for reporting this!
** 1997.11.07 SEA  Added code to place each type of database "object" in a
**                 separate procedure/method, and added a constant to allow
**                 the user to specify that indices be created in a pass
**                 separate from the one that creates the table.
** 1997.12.07  SEA Added code to initialize local variables for each parameter
**                 in a parameterized view.  Note that this only works if you
**                 have added each parameter to the view's parameters list,
**                 either through the View Designer (Query/View Parameters) or
**                 with DBSetProp().  Thanks to Todd Arnold for telling me how
**                 to create the parameters list!
** 1997.12.07  SEA Added code to display a dialog, warning the user that
**                 his/her dbc and/or tables are about to be overwritten, and
**                 requesting confirmation to continue.
** 1997.12.14  SEA Set EXACT ON in generated program so that the ASCAN() in 
**                 RemoveTable() won't find "TABLE10" when it's looking for
**                 "TABLE1".  Thanks to Jeff Clauss and Ken Lee for catching
**                 this!
** 1997.12.29  SEA Added code to break up the RI regeneration code into
**                 multiple procedures when it exceeds 64K.  Thanks to Arnon
**                 Gal-Oz for letting me know that it was time to do this! 
** 1998.07.20  SEA Fixed bug that created a procedure without a name if the
**                 database has no tables.
** 1998.07.20  SEA UPPER()'d and ALLTRIM()'d the dbc's ObjectName field when
**                 searching for relations in procedure GetRI.  Thanks to
**                 Patrick Godfrey for catching this!
** 2008.05.29  JME personnalisation
** 2009.01.18  JME plus de personnalisations
** 2009.02.09  JME ajout gnration classe d'accs
*******************************************************************************
LPARAMETERS tcOutFile, tcClassName, tlSeparateSPObjCode, tllaissenombdd
* jme 18/02/2009 tllaissenombdd si .T. on laisse le nom de la bdd dans les
*                clauses FROM des vues. .F. par dfaut : cela permet de
*                gnrer des vues qui seront stockes dans une autre bbd 
*                que celle utilise par gendbcxm (si on veut ajouter une
*                on la met dans une BDD 'temporaire', on gnre le programme
*                et on execute le programme gnr avec la bdd correcte.
#IF FILE("custproc.h")
	#INCLUDE custproc.h
#ENDIF

PRIVATE ALL EXCEPT g_*

**************************************************************************
** Constants
**************************************************************************
#DEFINE CRLF CHR(13) + CHR(10)
#DEFINE TAB  CHR(9)

*! SEA: CREATE TABLE command can exceed max command line length when 
*!      table has 180+ fields in 5.0, or 45+ fields in 3.0.  I'll try to
*!      alleviate the problem by allowing you to strip out the default
*!      formatting tabs and spaces.  If that's not enough, procedure
*!      GetTable will truncate the CREATE TABLE command and add as many
*!      ALTER TABLE commands as necessary to add the remaining fields.
#IFNDEF GDX_INDENT
	#DEFINE GDX_INDENT                   TAB
#ENDIF
#IFNDEF GDX_PRCMARGIN
	#DEFINE GDX_PRCMARGIN                GDX_INDENT
#ENDIF
#IFNDEF GDX_CMDMARGIN
	#DEFINE GDX_CMDMARGIN                GDX_PRCMARGIN + GDX_INDENT
#ENDIF
#IFNDEF GDX_FLDNAMEPAD
	#DEFINE GDX_FLDNAMEPAD               20
#ENDIF
#IFNDEF ALERT_USING_ALTER_TABLE
	#DEFINE ALERT_USING_ALTER_TABLE      .T.
#ENDIF
#IFNDEF ALTER_TABLE_NOVALIDATE
	#DEFINE ALTER_TABLE_NOVALIDATE       .F.
#ENDIF	

*! SEA: If you want to separate the index-creation routine from that which
*!      creates the related table, then #DEFINE MAKE_INDEX_PASS .T. in the
*!      CustProc.h header file.
#IFNDEF MAKE_INDEX_PASS
	#DEFINE MAKE_INDEX_PASS              .F.
#ENDIF

*! SEA: If you want to add field-level properties, such as CHECK, ERROR
*!      and DEFAULT separately from the code that creates the table, then
*!      #DEFINE CREATE_TABLE_SET_PROPS .F. in the CustProc.h header file.
#IFNDEF CREATE_TABLE_SET_PROPS
	#DEFINE CREATE_TABLE_SET_PROPS       .T.
#ENDIF

*! SEA: If you want your tables, views, connections and relations to be
*!      created in alphabetical order (rather than the order that they
*!      were created in the database container), then #DEFINE 
*!      GDX_SORT_OUTPUT .T. in the CustProc.h header file.
#IFNDEF GDX_SORT_OUTPUT
	#DEFINE GDX_SORT_OUTPUT              .F.
#ENDIF

**************************************************************************
** Error Messages
**************************************************************************
#DEFINE NO_DATABASE_IN_USE_LOC   "No Database is in use. " + ;
                                 "This program must have a database " + ;
                                 "available."
#DEFINE INVALID_PARAMETERS_LOC   "Invalid Parameters..." + CRLF + ;
                                 "An output file must be specified." + CRLF +;
                                 'ie: DO GENDBCX WITH "filename.prg"'
#DEFINE INVALID_DESTINATION_LOC  "Invalid Destination File "
#DEFINE NO_TEMP_FILE_LOC         "Could not create temporary file: "
#DEFINE NO_OUTPUT_WRITTEN_LOC    "Could not create or write to output file"
#DEFINE ERROR_TITLE_LOC          "Aborting GenDBCX..."
#DEFINE UNRECOVERABLE_LOC        "Unrecoverable Error: "
#DEFINE AT_LINE_LOC              " At Line: "
#DEFINE NO_FIND_LOC              "Could not set RI Information"
#DEFINE NO_FILE_FOUND_LOC        "Warning! No Procedure File Found!"
#DEFINE GETFILE_GEN_LOC          "Generate..."
#DEFINE NOT_SUPPORTED_LOC        "Filters on PRIMARY keys are not supported at this time. " + ;
                                 "A comment will be added to your output file specifying the filters."
#DEFINE NS_COMMENT_LOC           "****** These filters need to be added manually ******"
#DEFINE WARNING_TITLE_LOC        "GenDBCX Warning..." 
#DEFINE USING_ALTER_TABLE_LOC    "CREATE TABLE command line length exceeded, using" + CRLF + ;
                                 "ALTER TABLE to add remaining fields for table "
#DEFINE WARNING_OVERWRITING_LOC  "Warning!  This program will overwrite your existing database and tables!  Continue?"
#DEFINE PROGRAM_CANCELLED_LOC    "Program cancelled"

**************************************************************************
** Comments And Other Information
**************************************************************************
#DEFINE BEGIN_RELATION_LOC       "Relations Setup"
#DEFINE BEGIN_TABLE_LOC          "Table setup for "
#DEFINE BEGIN_INDEX_LOC          "Create each index for "
#DEFINE BEGIN_PROP_LOC           "Change properties for "
#DEFINE BEGIN_VIEW_LOC           "View setup for "
#DEFINE BEGIN_PROC_LOC           "Procedure Re-Creation"
#DEFINE BEGIN_CONNECTIONS_LOC    "Connection Definition"
#DEFINE BEGIN_RI_LOC             "Referential Integrity Setup"
#DEFINE OPEN_DATABASE_LOC        "Select Database..."
#DEFINE PACK_DATABASE_LOC        "Pack the Database"
#DEFINE SAVE_PRG_NAME_LOC        "Enter output program name..."
#DEFINE NO_MODIFY_LOC            "*** WARNING *** DO NOT MODIFY THIS FILE IN ANY WAY! *** WARNING ***"
#DEFINE TABLE_NAME_LOC           "*        Table Name: "
#DEFINE PRIMARY_KEY_LOC          "*       Primary Key: "
#DEFINE FILTER_EXP_LOC           "* Filter Expression: "
#DEFINE HEADING_LINE             "* *********************************************************************"
#DEFINE HEADING_1_LOC            HEADING_LINE + CRLF + ;
                                 "* *" + CRLF
#DEFINE HEADING_2_LOC            "* *" + CRLF + ;
                                 HEADING_LINE + CRLF + ;
                                 "* *" + CRLF + ;
                                 "* * Description:" + CRLF + ;
                                 "* * This program was automatically generated by GenDBCX Version 1.1," + CRLF + ;
                                 "* * a modified version of Microsoft's utility GenDBC Version 2.26.67." + CRLF + ;
                                 "* *" + CRLF + ;
                                 HEADING_LINE + CRLF

#DEFINE CREATE_DBC_LOC           "*-- Create the Database Container"
#DEFINE CREATE_TABLES_LOC        "*-- Create Tables"
#DEFINE CREATE_INDICES_LOC       "*-- Create Indices"
#DEFINE CREATE_VIEWS_LOC         "*-- Create Views"
#DEFINE CREATE_CONNECTIONS_LOC   "*-- Create Connections"
#DEFINE CALL_CUSTOM_LOC          "*-- Call custom method"
#DEFINE FINISH_TABLES_LOC        "*-- Finish Tables (add rules/triggers/properties)"

**************************************************************************
** Public Variables
**************************************************************************
IF SET("TALK") = "ON"                     && To restore SET TALK after use
	SET TALK OFF                           && -- Have to do it this way so
	g_cSetTalk = "ON"                      && -- nothing gets on screen
ELSE
	g_cSetTalk = "OFF"
ENDIF
g_cFullPath = SET("FULLPATH")             && To restore old FULLPATH setting
g_cOnError = ON("ERROR")                  && To restore old ON ERROR condition
g_cSetDeleted = SET("DELETED")            && To restore SET DELETED later
g_cSetStatusBar = SET("STATUS BAR")       && To restore STATUS bar
g_cStatusText = SYS(2001, "MESSAGE", 1)   && To restore text that may be on it
g_nMax = 7                                && For status line information
g_nCurrentStat = 1                        && For status line information
g_cFilterExp = ""                         && For Non-Supported Filter Info
g_cSetSafety = SET("SAFETY")              && To restore SAFETY setting
g_cSetExact  = SET("EXACT")               && To restore EXACT setting
g_project_window = WEXIST("Project Manager") && pour restaurer la fentre du gestionnaire de projet
**************************************************************************

*! Our generic error handling routine
ON ERROR DO GenDBC_Error WITH MESSAGE(), LINENO()

DEACTIVATE WINDOW "Project Manager"

*! Begin program body
SET DELETED ON
SET FULLPATH ON
IF m.g_cSetStatusBar = "OFF"
	SET STATUS BAR ON
ENDIF
SET SAFETY OFF
SET EXACT ON

*! Hook for a custom process to run after the DBC and table structures are created
lcHookStr = CALL_CUSTOM_LOC + CRLF + GDX_CMDMARGIN + "=this.CustomProcess()"

*! Make sure a database is open
IF EMPTY(DBC())
	g_cFullDatabase = GETFILE("DBC", OPEN_DATABASE_LOC, GETFILE_GEN_LOC, 0)
	IF EMPTY(m.g_cFullDatabase)
		=FatalAlert(NO_DATABASE_IN_USE_LOC, .F.)
	ENDIF
	OPEN DATABASE (m.g_cFullDatabase)
ENDIF

*! Set global variable to the database name and format it
g_cDatabase = FULLPATH(DBC())
IF RAT("\", m.g_cDatabase) > 0
	m.g_cDatabase = SUBSTR(m.g_cDatabase, RAT("\", m.g_cDatabase) + 1)
ENDIF

*! Get the fullpath of database
g_cFullDatabase = DBC()

*! Check for valid parameters
IF TYPE('m.tcOutFile') != "C" or EMPTY(m.tcOutFile)
	tcOutFile = ""
	tcOutFile = PUTFILE(SAVE_PRG_NAME_LOC, SUBSTR(m.g_cDatabase, 1, RAT(".", m.g_cDatabase)) + "PRG", "PRG")
	IF EMPTY(m.tcOutFile)
		=FatalAlert(INVALID_PARAMETERS_LOC, .F.)
	ENDIF
ENDIF

*! Check for proper extensions or add one if none specified
IF RAT(".PRG", m.tcOutFile) = 0 AND RAT(".", m.tcOutFile) = 0
	tcOutFile = m.tcOutFile + ".PRG"
ENDIF

*! Make sure the output file is valid
hFile = FCREATE(m.tcOutFile)
IF m.hFile <= 0
	=FatalAlert(INVALID_DESTINATION_LOC + m.tcOutFile, .F.)
ENDIF
=FCLOSE(m.hFile)
ERASE (m.tcOutFile)

*! Remember all our tables that are open for this database
g_nTotal_Tables_Used = AUSED(g_aAlias_Used)
IF m.g_nTotal_Tables_Used > 0
	DIMENSION g_aTables_Used[m.g_nTotal_Tables_Used]

	*! Get real names of tables opened
	FOR nLoop = 1 TO m.g_nTotal_Tables_Used
		g_aTables_Used[m.nLoop] = DBF(g_aAlias_Used[m.nLoop, 1])
	ENDFOR
ENDIF

*! Get number of tables contained in database
nTotal_Tables = ADBOBJECTS(aAll_Tables, "Table")
g_nMax = m.g_nMax + m.nTotal_Tables
=Stat_Message()
* jme ?


*! Get number of views contained in database
nTotal_Views = ADBOBJECTS(aAll_Views, "View")
DIMENSION g_ltsqlvue( m.nTotal_Views) && jme 09/02/2009
g_ltsqlvue = ""
g_nMax = m.g_nMax + m.nTotal_Views
=Stat_Message()

*! Get number of connections contained in database
nTotal_Connections = ADBOBJECTS(aAll_Connections, "Connection")
g_nMax = m.g_nMax + m.nTotal_Connections
=Stat_Message()

*! Get number of relations contained in database
nTotal_Relations = ADBOBJECTS(aAll_Relations, "Relation")
*! SEA: We process relations so fast now, it's better not to include them
*! in the "status" line, as they distort the progress indicator
* m.g_nMax = m.g_nMax + m.nTotal_Relations
=Stat_Message()

CLOSE DATABASES ALL

**************************
*** Get Stored Procedures
**************************
*! Create an output file that will be appended to the database
*! as procedures
cFile = UPPER(SUBSTR(m.tcOutFile, 1, AT(".", m.tcOutFile))) + "krt"
cDPTemp = "GenDBC.$$$"
cSPTemp = "GenDBC.k$$"
IF tlSeparateSPObjCode
	cSPObjTemp = "GenDBC.o$$"
	cSPObjFile = UPPER(SUBSTR(m.tcOutFile, 1, AT(".", m.tcOutFile))) + "kro"
ENDIF

*! Place Header Information For Source/Object
hFile = FCREATE(m.cSPTemp)
IF m.hFile <= 0
	=FCLOSE(m.hFile)
	=FatalAlert(NO_OUTPUT_WRITTEN_LOC, .T.)
ENDIF

=FPUTS(m.hFile, NO_MODIFY_LOC)
=FCLOSE(m.hFile)

*! Now we are going to copy the object and source code
*! for the stored procedures
OPEN DATABASE (m.g_cFullDatabase) EXCLUSIVE
COMPILE DATABASE (m.g_cFullDatabase)
CLOSE DATABASES ALL
USE (m.g_cFullDatabase) SHARED
LOCATE FOR Objectname = 'StoredProceduresSource'
IF FOUND()
	COPY MEMO Code TO (m.cSPTemp) ADDITIVE
ENDIF
=ADIR(aTemp, m.cSPTemp)
nSourceSize = m.aTemp[1,2] - LEN(NO_MODIFY_LOC)

LOCATE FOR Objectname = 'StoredProceduresObject'
IF FOUND()
	IF tlSeparateSPObjCode
		COPY MEMO Code TO (m.cSPObjTemp)
	ELSE
		COPY MEMO Code TO (m.cSPTemp) ADDITIVE
	ENDIF
ENDIF
USE

*! Open the database again
OPEN DATABASE (m.g_cFullDatabase) EXCLUSIVE

hOutFile = FCREATE(cDPTemp)
IF m.hOutFile <= 0
	=FCLOSE(m.hFile)
	=FatalAlert(NO_OUTPUT_WRITTEN_LOC, .T.)
ENDIF

*******************************************************************************
** Create a class that holds all of the database regeneration code as
** public methods that can be called by anyone who can create a reference
** to the class.
**
** The body of the program now consists of two lines: 1) the creation of
** the GenDBCX class object, and 2) a call to a "summary" method that 
** calls all of the methods that create the database, tables, views and 
** connections, and set relations and referential integrity.
**
** Using this approach, you can now DO the program to recreate the entire
** database, or you can create an object reference to the GenDBCX class,
** and recreate tables, views and connections by appending the table/view
** name to its related prefix ("tb" for tables, "vw" for views, "cn" for
** connections), relations by calling the SetRelations method, and RI by
** calling the SetRI method.
** 
** Additionally, if you want to execute the code of a method without
** creating a GenDBCX object, you can call the program with the name of
** a single method to execute.
*******************************************************************************
*! Write the program heading
=WriteFile(m.hOutFile, ;
	HEADING_1_LOC + "* * " + DTOC(DATE()) +;
	SPACE(25 - LEN(m.g_cDatabase) / 2) + ;
	m.g_cDatabase + SPACE(25 - LEN(m.g_cDatabase) / 2) +;
	TIME() + CRLF + HEADING_2_LOC + CRLF + ;
	IIF(!EMPTY(m.g_cFilterExp), NS_COMMENT_LOC +  m.g_cFilterExp + ;
	CRLF + REPLICATE("*", 52) + CRLF, ""))

*! Write the body of the program
lcObjectName = "oDBC"
lcClassName = IIF(TYPE("m.tcClassName")="C" AND !EMPTY(m.tcClassName),tcClassName,"GenDBCX")
lcGenAllProc = "GenerateAll"
=WriteFile(m.hOutFile, "* ce code ne pourra pas tre excut s'il y a une ou plusieurs")
=WriteFile(m.hOutFile, "* classe d'accs aux tables ci-dessus.")
=WriteFile(m.hOutFile, ;
	"LPARAMETERS tcMethod" + CRLF + ;
	"LOCAL " + lcObjectName + CRLF + ;
	lcObjectName + " = CreateObject('" + m.lcClassName + "')" + CRLF + ;
	               "IF TYPE('" + lcObjectName + "') != 'O'" + CRLF + ;
	GDX_INDENT +   	"RETURN" + CRLF + ;
	               "ENDIF" + CRLF + ;
	               "cExactSet = SET('EXACT')" + CRLF + ;
	               "SET EXACT ON" + CRLF + ;
	               "IF EMPTY(tcMethod)" + CRLF + ;
	GDX_INDENT +   	"=" + lcObjectName + "." + m.lcGenAllProc + "()" + CRLF + ;
	               "ELSE" + CRLF + ;
	GDX_INDENT +   	"=EVALUATE('" + lcObjectName + ".' + tcMethod)" + CRLF + ;
	               "ENDIF" + CRLF + ;
	               "IF cExactSet = 'OFF'" + CRLF + ;
	GDX_INDENT +   	"SET EXACT OFF" + CRLF + ;
	               "ENDIF" + CRLF + CRLF ;
)

*! Begin defining the class
=WriteFile(m.hOutFile, ;
	"DEFINE CLASS " + m.lcClassName + " AS Custom" + CRLF + ;
	GDX_PRCMARGIN + "PROTECTED aTables[1]" + CRLF + ;
	GDX_PRCMARGIN + "PROTECTED nTables" + CRLF ;
)

*! Write the CustomProcess method
#IFDEF CUSTOM_METHOD_LOC
	=WriteFile(m.hOutFile, ;
		GDX_PRCMARGIN + "PROCEDURE CustomProcess" + CRLF + ;
		GDX_CMDMARGIN +    CUSTOM_METHOD_LOC + ;
		GDX_PRCMARGIN + "ENDPROC" + CRLF ;
	)
#ELSE
	=WriteFile(m.hOutFile, ;
		GDX_PRCMARGIN + "PROCEDURE CustomProcess" + CRLF + ;
		GDX_CMDMARGIN +    "*-- Your code goes here" + CRLF + ;
		GDX_PRCMARGIN + "ENDPROC" + CRLF ;
	)
#ENDIF

*! Write the Init method, storing the names of any tables in an existing dbc
*! into the class' "table names" array
=WriteFile(m.hOutFile, ;
	GDX_PRCMARGIN +              "PROCEDURE Init" + CRLF + ;
	GDX_CMDMARGIN +                 "IF FILE('" + m.g_cDatabase + "')" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +       "OPEN DATABASE " + m.g_cDatabase + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +       "this.nTables = ADBOBJECTS(this.aTables, 'Table')" + CRLF + ;
	GDX_CMDMARGIN +                 "ENDIF" + CRLF + ;
	GDX_CMDMARGIN +                 "LOCAL ARRAY xx[1]" + CRLF + ;
	GDX_CMDMARGIN +                 "IF ADIR(xx,'*.DBC') > 0 OR ADIR(xx,'*.DBF') > 0" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +       "IF MessageBox('" + WARNING_OVERWRITING_LOC + "', " + ;
	                                       LTRIM(STR(48+4+256)) + ;
	                                       ", '" + WARNING_TITLE_LOC + "') != 6" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT + "RETURN .F." + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +       "ENDIF" + CRLF + ;
	GDX_CMDMARGIN +                 "ENDIF" + CRLF + ;
	GDX_PRCMARGIN +              "ENDPROC" + CRLF ;
)

*! Create the call to the DBC recreation method, to be inserted into the
*! "summary" method below
lcCreateDBCProc = "CreateDBC"
lcCDStr = "=this." + m.lcCreateDBCProc + "()"

*! Begin the database container recreation method
=WriteFile(m.hOutFile, ;
	GDX_PRCMARGIN + "PROCEDURE " + m.lcCreateDBCProc + CRLF + ;
	GDX_CMDMARGIN +    CREATE_DBC_LOC + CRLF + ;
	GDX_CMDMARGIN +    "CLOSE DATA ALL" + CRLF + ;
	GDX_CMDMARGIN +    "CREATE DATABASE '" + m.g_cDatabase + "'" ;
)

*! If the database contained any stored procedures (i.e. we wrote something
*! other than the NO_MODIFY_LOC constant to the .k$$ file), then output
*! the code required to recreate the procedure file.
=ADIR(aTemp, m.cSPTemp)
IF m.aTemp[1, 2] > LEN(NO_MODIFY_LOC) + LEN(CRLF)
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "********* " + BEGIN_PROC_LOC + " *********")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "IF !FILE([" + SUBSTR(m.cFile, RAT("\", m.cFile) + 1) + "])")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "? [" + NO_FILE_FOUND_LOC + "]")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "ELSE")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "CLOSE DATABASE")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "USE '" +  m.g_cDatabase + "'")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "g_SetSafety = SET('SAFETY')")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "SET SAFETY OFF")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "LOCATE FOR Objectname = 'StoredProceduresSource'")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "IF FOUND()")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +        "APPEND MEMO Code FROM [" + SUBSTR(m.cFile, RAT("\", m.cFile) + 1) + "] OVERWRITE")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +        "REPLACE Code WITH SUBSTR(Code, " + ALLTRIM(STR(LEN(NO_MODIFY_LOC) + 3)) + ", " + ALLTRIM(STR(m.nSourceSize - 2)) + ")")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "ENDIF")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "LOCATE FOR Objectname = 'StoredProceduresObject'")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "IF FOUND()")
	IF tlSeparateSPObjCode
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +        "APPEND MEMO Code FROM [" + SUBSTR(m.cSPObjFile, RAT("\", m.cSPObjFile) + 1) + "] OVERWRITE")
	ELSE
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +        "APPEND MEMO Code FROM [" + SUBSTR(m.cFile, RAT("\", m.cFile) + 1) + "] OVERWRITE")
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +        "REPLACE Code WITH SUBSTR(Code, " + ALLTRIM(STR(LEN(NO_MODIFY_LOC) + m.nSourceSize + 1)) + ")")
	ENDIF
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "ENDIF")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "IF UPPER(g_SetSafety) = 'ON'")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +        "SET SAFETY ON")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "ENDIF")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "USE")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "OPEN DATABASE [" + m.g_cDatabase + "]")
	=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "ENDIF")
	=WriteFile(m.hOutFile, "")
ELSE
	ERASE (m.cSPTemp)
ENDIF

*! Close the database container recreation method
=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC" + CRLF)

=Stat_Message()

**************************
*** Get Tables
**************************
STORE "" TO lcCTProc, lcCRTPProc, lcCIDXProc, lcCrtTables, lcCrtIndices, lcCrtRTPs, ;
 lcCTStr, lcCRTPStr, lcCIDXStr

lcProcPrefix = "tb"
IF m.nTotal_Tables > 0
	lcCTProc     = "CreateTables"
	lcCIDXProc   = "CreateIndices"
	lcCRTPProc   = "RulesTriggersProperties"
	lcCrtTables  = "=this." + lcCTProc + "()"
	lcCrtIndices = "=this." + lcCIDXProc + "()"
	lcCrtRTPs    = "=this." + lcCRTPProc + "()"
	lcCTStr      = GDX_CMDMARGIN + CREATE_TABLES_LOC
	lcCIDXStr    = GDX_CMDMARGIN + CREATE_INDICES_LOC
	lcCRTPStr    = GDX_CMDMARGIN + FINISH_TABLES_LOC
	#IF GDX_SORT_OUTPUT
		=ASORT(aAll_Tables,1)
	#ENDIF
	FOR m.nLoop = 1 TO m.nTotal_Tables
		lcProcName = m.lcProcPrefix + ALLTRIM(CHRTRAN(aAll_Tables[m.nLoop]," ","_"))
		DO GetTable WITH ALLTRIM(aAll_Tables(m.nLoop)), m.hOutFile, m.lcProcName
		lcCTStr   = lcCTStr   + CRLF + GDX_CMDMARGIN + "=this." + m.lcProcName + "(1)"
		lcCRTPStr = lcCRTPStr + CRLF + GDX_CMDMARGIN + "=this." + m.lcProcName + "(2)"
		lcCIDXStr = lcCIDXStr + CRLF + GDX_CMDMARGIN + "=this." + m.lcProcName + "(3)"
		=Stat_Message()
	ENDFOR
ENDIF

**************************
*** Get Connections
**************************
STORE "" TO lcCCProc, lcCrtConns, lcCCStr

lcProcPrefix = "cn"
IF m.nTotal_Connections > 0
	lcCCProc = "CreateConnections"
	lcCrtConns = "=this." + lcCCProc + "()"
	lcCCStr = GDX_CMDMARGIN + CREATE_CONNECTIONS_LOC
	#IF GDX_SORT_OUTPUT
		=ASORT(aAll_Connections,1)
	#ENDIF
	FOR m.nLoop = 1 TO m.nTotal_Connections
		lcProcName = m.lcProcPrefix + ALLTRIM(CHRTRAN(aAll_Connections[m.nLoop]," ","_"))
		DO GetConn WITH ALLTRIM(aAll_Connections[m.nLoop]), m.hOutFile, m.lcProcName
		lcCCStr = lcCCStr + CRLF + GDX_CMDMARGIN + "=this." + m.lcProcName + "()"
		=Stat_Message()
	ENDFOR
ENDIF

**************************
*** Get Views
**************************
STORE "" TO lcCVProc, lcCrtViews, lcCVStr
lcProcPrefix = "vw"
IF m.nTotal_Views > 0
	lcCVProc = "CreateViews"
	lcCrtViews = "=this." + m.lcCVProc + "()"
	lcCVStr = GDX_CMDMARGIN + CREATE_VIEWS_LOC
	#IF GDX_SORT_OUTPUT
		=ASORT(aAll_Views,1)
	#ENDIF
	FOR m.nLoop = 1 TO m.nTotal_Views
		lcProcName = m.lcProcPrefix + ALLTRIM(CHRTRAN(aAll_Views[m.nLoop]," ","_"))
		DO GetView WITH ALLTRIM(aAll_Views[m.nLoop]), m.hOutFile, m.lcProcName, m.nloop, m.tllaissenombdd
		lcCVStr = lcCVStr + CRLF + GDX_CMDMARGIN + "=this." + m.lcProcName + "()"
		=Stat_Message()
	ENDFOR
ENDIF

**************************
*** Get Relations
**************************
lcSetRelStr = ""
lcProcName = "SetRelations"
IF m.nTotal_Relations > 0
	lcSetRelStr = "=this." + m.lcProcName + "()"
	LOCAL lcRelString
	lcRelString = CRLF + GDX_PRCMARGIN + "***** " + BEGIN_RELATION_LOC + " *****" + ;
	              CRLF + GDX_PRCMARGIN + "PROCEDURE " + m.lcProcName + CRLF
	#IF GDX_SORT_OUTPUT
		=ASORT(aAll_Relations,1)
	#ENDIF
	FOR m.nLoop = 1 TO m.nTotal_Relations
		lcRelString = lcRelString + GDX_CMDMARGIN + "ALTER TABLE '" + aAll_Relations[m.nLoop,1] +;
							 "' ADD FOREIGN KEY TAG " + aAll_Relations[m.nLoop,3] + ;
							 " REFERENCES " + IIF(" " $ aAll_Relations[m.nLoop,2], ["], []) + ;
							    aAll_Relations[m.nLoop,2] + IIF(" " $ aAll_Relations[m.nLoop,2], ["], []) + ;
							 " TAG " + aAll_Relations[m.nLoop,4] + IIF(ALTER_TABLE_NOVALIDATE, " NOVALIDATE", "") + CRLF
		*! SEA: As mentioned above, we process relations so fast now, it's better 
		*! not to include them in the progress bar, as they distort the indicator
		* =Stat_Message()
	ENDFOR
	lcRelString = m.lcRelString + GDX_PRCMARGIN + "ENDPROC" + CRLF
	=WriteFile(m.hOutFile, m.lcRelString)
ENDIF

CLOSE DATABASE  && Because we're going to start peeking into the
                && table structure of the DBC

**************************
*** Get RI Info
**************************
lcSetRIStr = ""
lcProcName = "SetRI"
IF m.nTotal_Relations > 0
	DO GetRI WITH m.hOutFile, m.lcProcName
	lcSetRIStr = "=this." + m.lcProcName + "()"
ENDIF
=Stat_Message()

******************************************************************************
*! Create the RemoveTable() function, which allows individual methods to
*! recreate a table that is part of the database, but has been accidentally
*! erased.
******************************************************************************
=WriteFile(m.hOutFile, CRLF + ;
	GDX_PRCMARGIN +                           "FUNCTION RemoveTable(tcTable, tcFile)" + CRLF + ;
	GDX_CMDMARGIN +                              "LOCAL ARRAY laTables[1]" + CRLF + ;
	GDX_CMDMARGIN +                              "=ADBOBJECTS(laTables,'Table')" + CRLF + ;
	GDX_CMDMARGIN +                              "IF ASCAN(laTables,tcTable) = 0" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- No such table in the current dbc, nothing to do!" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "RETURN .t." + CRLF + ;
	GDX_CMDMARGIN +                              "ENDIF" + CRLF + ;
	GDX_CMDMARGIN +                              "IF !FILE(tcFile)" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- If the file has been erased, REMOVE TABLE fails with a 'File does not" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- exist' error, because it can't find the file to either remove its" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- database reference or delete the file ." + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- We'll work around this problem by creating a phony file with the same" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- name, and use REMOVE TABLE DELETE to both remove the table reference" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "*-- in the dbc, and erase the phony file." + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "LOCAL lhFile" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "lhFile = FCREATE(tcFile)" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "IF lhFile = -1" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +          "RETURN .f." + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "ENDIF" + CRLF + ;
	GDX_CMDMARGIN + GDX_INDENT +                    "=FCLOSE(lhFile)" + CRLF + ;
	GDX_CMDMARGIN +                              "ENDIF" + CRLF + ;
	GDX_CMDMARGIN +                              "REMOVE TABLE (tcTable) DELETE" + CRLF + ;
	GDX_CMDMARGIN +                              "=ADBOBJECTS(laTables,'Table')" + CRLF + ;
	GDX_CMDMARGIN +                              "RETURN ASCAN(laTables,tcTable) = 0" + CRLF + ;
	GDX_PRCMARGIN +                           "ENDFUNC" + CRLF ;
)

******************************************************************************
*! Create the "summary" procedure that calls all the other procedures
*! to create tables, views, etc.
******************************************************************************
=WriteFile(m.hOutFile, "")
=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + lcGenAllProc)
=WriteFile(m.hOutFile, GDX_CMDMARGIN +    m.lcCDStr)
IF !EMPTY(m.lcCrtTables)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + m.lcCrtTables)
ENDIF
=WriteFile(m.hOutFile, "")
=WriteFile(m.hOutFile, GDX_CMDMARGIN +    m.lcHookStr)
=WriteFile(m.hOutFile, "")
IF MAKE_INDEX_PASS and !EMPTY(m.lcCrtIndices)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + m.lcCrtIndices)
ENDIF
=WriteFile(m.hOutFile, GDX_CMDMARGIN +    m.lcCrtRTPs)
IF !EMPTY(m.lcCrtConns)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + m.lcCrtConns)
ENDIF
IF !EMPTY(m.lcCrtViews)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + m.lcCrtViews)
ENDIF
IF !EMPTY(m.lcSetRelStr)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + m.lcSetRelStr)
ENDIF
IF !EMPTY(m.lcSetRIStr)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + m.lcSetRIStr)
ENDIF
=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")

IF !EMPTY(m.lcCTStr)
******************************************************************************
*! Create a method that creates the tables in the database
******************************************************************************
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.lcCTProc)
	=WriteFile(m.hOutFile,                     m.lcCTStr)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
ENDIF

IF MAKE_INDEX_PASS and !EMPTY(m.lcCIDXStr)
******************************************************************************
*! Create a method that creates the indexes for each table
******************************************************************************
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.lcCIDXProc)
	=WriteFile(m.hOutFile,                     m.lcCIDXStr)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
ENDIF

IF !EMPTY(m.lcCRTPStr)
******************************************************************************
*! Create a method that creates the rules, triggers and properties for each table
******************************************************************************
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.lcCRTPProc)
	=WriteFile(m.hOutFile,                     m.lcCRTPStr)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
ENDIF

IF !EMPTY(m.lcCCStr)
******************************************************************************
*! Create a method that creates the database's connections
******************************************************************************
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.lcCCProc)
	=WriteFile(m.hOutFile,                     m.lcCCStr)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
ENDIF

IF !EMPTY(m.lcCVStr)
******************************************************************************
*! Create a method that creates the database's views
******************************************************************************
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.lcCVProc)
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  PRIVATE lcv, lcf, lcp, lckf, lcupd, lcupn, lcrle, lcrlt, ;") && jme 29/05/2008
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "    lccap, lccom, lcdt, lcdfv, lcvue")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  PRIVATE lcvudt, lcvwt, lcvfm, lcvsu, lcvums, lcvfs, lcvmr, ;") && jme 29/05/2008
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "    lcvt, lcvcom, lcvbuc, lcvsc, lcvp, lcvcm, lcvfan, lcvpl ")
  =WriteFile(m.hOutFile, "")
*!*    =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  STORE '' TO lcv, lcf, lcp") jme 01/02/2009
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  STORE '' TO lcp, lcvue")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcv = 'View'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcf = 'Field'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lckf = 'KeyField'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcupd = 'Updatable'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcupn = 'UpdateName'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcrle = 'RuleExpression'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcrlt = 'RuleText'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lccap = 'Caption'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lccom = 'Comment'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcdt = 'DataType'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcdfv = 'DefaultValue'")
*  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "   = )
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvudt = 'UpdateType'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvwt = 'WhereType'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvfm = 'FetchMemo'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvsu = 'SendUpdates'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvums = 'UseMemoSize'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvfs = 'FetchSize'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvmr = 'MaxRecords'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvt = 'Tables'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvcom = 'Comment'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvbuc = 'BatchUpdateCount'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvsc = 'ShareConnection'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvp = 'Prepared'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvcm = 'CompareMemo'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvfan = 'FetchAsNeeded'")
  =WriteFile(m.hOutFile, GDX_PRCMARGIN + "  lcvpl = 'ParameterList'")
*  #############
  
	=WriteFile(m.hOutFile,                     m.lcCVStr)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
ENDIF
******************************************************************************
*! End the class definition
******************************************************************************
=WriteFile(m.hOutFile, "ENDDEFINE &"+ "& "+m.lcClassName)
=FCLOSE(m.hOutFile)

* jme 09/02/2009 on cre la classe d'accs aux donnes pour chaque table
* S'il y a une vue _RW sur la table, on la met  jour via cette vue donc
* avec un tableupdate.
* S'il n'y a pas de vue _RW, on mete  jour la table avec INSERT INTO, 
* GATHER NAME, ...
IF m.nTotal_Tables > 0
  LOCAL lcnomfichtmp AS String, lnaccu AS Number
  lcnomfichtmp = creation_classe_acces()
  IF !EMPTY( m.lcnomfichtmp)
    lnaccu = STRTOFILE( FILETOSTR(m.lcnomfichtmp), m.tcOutFile, 0)
    lnaccu = STRTOFILE( FILETOSTR(m.cDPTemp), m.tcOutFile, 1)
    * on n'efface pas le fichier temporaire, il sera cras  la prochaine gnration
  ELSE
    COPY FILE (m.cDPTemp) TO (m.tcOutFile)
  ENDIF && !EMPTY( m.lcnomfichtmp)
ELSE
  COPY FILE (m.cDPTemp) TO (m.tcOutFile)
ENDIF && m.nTotal_Tables > 0


******************************************************************************
*! Make the output file(s) permanent
******************************************************************************
*COPY FILE (m.cDPTemp) TO (m.tcOutFile)

ERASE (m.cDPTemp)
IF FILE(m.cSPTemp)
	COPY FILE (m.cSPTemp) to (m.cFile)
	ERASE (m.cSPTemp)
ENDIF
IF tlSeparateSPObjCode AND FILE(m.cSPObjTemp)
	COPY FILE (m.cSPObjTemp) to (m.cSPObjFile)
	ERASE (m.cSPObjTemp)
ENDIF

=Stat_Message()

COMPILE (m.tcOutFile)

=GenDBC_CleanUp(.T.)

*!*  IF m.g_project_window && jme 09/02/2009
*!*    * le gestionnaire de projet tait ouvert, on le reouvre
*!*    ACTIVATE WINDOW "Project Manager" ne fonctionne pas
*!*  ENDIF && m.g_project_window

MODIFY FILE (m.tcOutFile) NOWAIT

*********************** END OF PROGRAM ***********************

**************************************************************************
**
** Function Name: GETRI()
** Creation Date: 1994.12.02
** Purpose:
**
**      To take existing FoxPro 3.0/5.0 RI Information, and generate an output
**      program that can be used to "re-create" this.
**
** Parameters:
**
**      hOutFile  - The file handle of the output file
**      cProcName - A character string containing the name of the
**                  current procedure
**
** Modification History:
**
**  1995.01.05  KRT  Created function
**  1995.12.20  KRT  Allow better lookup when trying to find
**                   the right record to add the RI information
**  1997.02.24  SEA  Output code within a PROCEDURE / ENDPROC block to
**                   avoid error when compiling procedures > 64K
**                   database re-creation program a procedure
**  1997.02.24  SEA  Write directly to a single output file, the handle 
**                   of which is passed in hOutFile
**  1997.12.29  SEA  Added code to break up the RI regeneration code
**                   into multiple procedures when it exceeds 64K
**  1998.03.26  SEA  ALLTRIM()'d the dbc's ObjectName field when searching
**                   for relations.
**************************************************************************
PROCEDURE GetRI
	LPARAMETERS hOutFile, cProcName

	PRIVATE ALL EXCEPT g_*

	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "***** " + BEGIN_RI_LOC + " *****")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + cProcName)

	*! USE the database
	USE (m.g_cFullDatabase) EXCLUSIVE

	LOCATE FOR ObjectType = "Relation" AND !EMPTY(RiInfo)
	IF FOUND()
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + "CLOSE DATABASE")
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + "USE '" +  m.g_cDatabase + "'")
		lnRILen = 50	&& approx length of code so far
		lnRIProcNum = 1
		DO WHILE FOUND()
			*! SEA: If we're about to exceed the max size of a procedure (64K),
			*! close this proc and start another one, which we'll call from this proc.
			*! Since we're using approximations, I'll cut off at 64000 instead of
			*! 65000+.
			IF lnRILen > 64000
				=WriteFile(m.hOutFile, "")
				=WriteFile(m.hOutFile, GDX_CMDMARGIN + "=" + cProcName+"_"+ltrim(str(lnRIProcNum))+"()")
				=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
				=WriteFile(m.hOutFile, "")
				=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE "+cProcName+"_"+ltrim(str(lnRIProcNum)))
				lnRIProcNum = lnRIProcNum + 1
				lnRILen = 20
			ENDIF
			*! Have to get the parent name to verify we are adding
			*! Information to the right record.
			m.nParentID = ParentID
			*! We use select so we won't mess up our LOCATE ... CONTINUE command
			SELECT ObjectName FROM (m.g_cFullDatabase) WHERE ObjectID = nParentID INTO ARRAY aTableName
			nStart = 1
			cITag = ""
			cTable = ""
			cRTag = ""
			DO WHILE m.nStart <= LEN(Property)
				nSize = ASC(SUBSTR(Property, m.nStart, 1)) +;
				(ASC(SUBSTR(Property, m.nStart + 1, 1)) * 256) +;
				(ASC(SUBSTR(Property, m.nStart + 2, 1)) * 256^2) + ;
				(ASC(SUBSTR(Property, m.nStart + 3, 1)) * 256^3)

				nKey = ASC(SUBSTR(Property, m.nStart + 6, 1))

				DO CASE
					CASE m.nKey = 13
						cITag = SUBSTR(Property, m.nStart + 7, m.nSize - 8)
					CASE m.nKey = 18
						cTable = SUBSTR(Property, m.nStart + 7, m.nSize - 8)
					CASE m.nKey = 19
						cRTag = SUBSTR(Property, m.nStart + 7, m.nSize - 8)
				ENDCASE
				nStart = m.nStart + m.nSize
			ENDDO
			=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "LOCATE FOR ObjectType = 'Table' AND UPPER(ALLTRIM(ObjectName)) == '" + UPPER(ALLTRIM(aTableName[1])) + "'")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "IF FOUND()")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "nObjectID = ObjectID")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "LOCATE FOR ObjectType = 'Relation' AND '" + m.cITag + ;
						                                                              "'$Property AND '" + m.cTable + "'$Property AND '" + m.cRTag + ;
						                                                              "'$Property AND ParentID = nObjectID")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "IF FOUND()")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +       "REPLACE RiInfo WITH '" + RiInfo + "'")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "ELSE")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +       '? "' + NO_FIND_LOC + ', line " + ltrim(str(lineno()))')
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "ENDIF")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN +                           "ENDIF")
			lnRILen = lnRILen + 400		&& approximate length of code we just added
			CONTINUE
		ENDDO
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + "USE")
	ENDIF
	USE
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
RETURN

**************************************************************************
**
** Function Name: GETTABLE(, )
** Creation Date: 1994.12.01
** Purpose        :
**
**              To take an existing FoxPro Table, and generate an output
**              program that can be used to "re-create" that Table.
**
** Parameters:
**
**      cTableName       A character string representing the name of the
**                       existing Table
**
**      hOutFile         The handle of the output file
**
**      cProcName        The name of the procedure within which we wrap the
**                       code that recreates the table
**
** Modification History:
**
**  1994.12.02  KRT      Created function
**  1994.12.05  KRT      Made it a function and cleaned it up
**  1994.12.08  KRT      Assume Database is open to speed up operation
**  1995.09.15  KRT      Use DBF() to find the real table name
**  1995.09.15  KRT      Take into account CR+LF in comment fields
**  1996.04.12  KRT      Added new properties for Visual FoxPro 5.0
**                       InputMask / Format / DisplayClass
**                       DisplayClassLibrary
**  1996.06.01  KRT      Added support for Collate sequence on index
**  1997.02.24  SEA      Output code within a PROCEDURE / ENDPROC block to
**                       avoid error when compiling procedures > 64K
**  1997.02.24  SEA      Write directly to a single output file, the handle 
**                       of which is passed in hOutFile
**  1997.02.25  SEA      Changed format of CREATE TABLE output code to
**                       improve its readability
**  1997.02.28  SEA      Only write the field ERROR clause if we have
**                       a CHECK clause
**  1997.05.01  SAS      Make two passes through the routine, the first to
**                       create the table, and the second to create its
**                       properties.
**  1997.09.05  SEA      When creating the procedure name, replace spaces
**                       with underscores.
**  1997.11.01  SEA      Added code to handle "Program line too long" error
**                       which occurs when a table has a large number of fields
**  1997.11.07  SEA      Added a constant to allow the user to specify that
**                       indices be created in a pass separate from the one
**                       that creates the table.
**  1997.12.16  SEA      Check for TAGCOUNT() > 0 before writing any index
**                       regeneration code.  Thanks to Mike Rothaus for
**                       suggesting this!
*******************************************************************************

PROCEDURE GetTable
	LPARAMETERS cTableName, hOutFile, cProcName

	PRIVATE ALL EXCEPT g_*

	*! Open Table to get field info
	USE (m.cTableName) EXCLUSIVE

	*! Get all the fields
	nNumberOfFields = AFIELDS(aAll_Fields)

	*! Header Information
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "**************************************************")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "** " + BEGIN_TABLE_LOC + m.cTableName)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "**************************************************")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.cProcName)
	=WriteFile(m.hOutFile, GDX_CMDMARGIN +    "LPARAMETERS tnPassNumber")
	=WriteFile(m.hOutFile, "")
	
	*! NOTE 
	*! If the table is greater than 8 characters then it will fail on platforms that
	*! do not support this (Such as Win32s).
	cOldSetFullPath = SET("FULLPATH")
	SET FULLPATH ON
	cTableFileName = DBF(ALIAS())

	*-- Start pass 1, create the table
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + "DO CASE" + CRLF + ;
	                       GDX_CMDMARGIN + "CASE tnPassNumber = 1")
		
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "=this.RemoveTable('" + m.cTableName + "', '" + ;
		m.cTableFileName + "')" + CRLF)

	SET FULLPATH OFF
	cTableFileName = DBF(ALIAS())
	IF UPPER(cOldSetFullPath) = "ON"
		SET FULLPATH ON
	ENDIF

	*! SEA: The CREATE TABLE command we construct below can exceed the command line
	*!      length limitation, especially in 3.0.  We can work around this problem
	*!      by using CREATE TABLE to add as many fields as possible, and adding the
	*!      rest through ALTER TABLE commands.
*!*  	IF "VISUAL FOXPRO 03" $ upper(version(1)) jme 18/01/2009
*!*  		nMaxCommandLen = 2048		&& max in 3.0 is 2K
*!*  	ELSE
		nMaxCommandLen = 8192		&& max in 5.0 or greater is (at least) 8K
*!*  	ENDIF
	nMaxLine = 100		&& an arbitrary buffer to ensure we don't exceed the limit
	lUseAlterTable = .F.

*!*  	cFieldLineEnd = ", ;" + CRLF jme 18/01/2009
*!*    cFieldLineEnd = ", ; && " + CRLF IL CROIT QUE C'EST UN COMMENTAIRE !!!!
*!* et il faut laisser cette ligne vide

  cFieldLineEnd = ", ; &" +"& " + CRLF
	cTableFileName = SUBSTR(m.cTableFileName, RAT(":", m.cTableFileName) + 1)
	cCreateTable = GDX_CMDMARGIN + GDX_INDENT + "CREATE TABLE '" + m.cTableFileName + ;
		"' NAME '" + m.cTableName + "' ( ;" + CRLF
	
	*! Information about each field that can been written with CREATE TABLE - SQL
	FOR nInner_Loop = 1 TO m.nNumberOfFields
		*! If we're approaching the command line length limit, switch to "ALTER TABLE"
		IF !lUseAlterTable and LEN(cCreateTable) > nMaxCommandLen - nMaxLine
			IF ALERT_USING_ALTER_TABLE
				=MessageBox(USING_ALTER_TABLE_LOC + m.cTableName, 64, WARNING_TITLE_LOC)
			ENDIF
			*! Since the previous field is the last we're including in the
			*! CREATE TABLE command, we have to remove the comma that follows
			*! the field info.
			cCreateTable = left(cCreateTable,len(cCreateTable)-len(cFieldLineEnd)) + " ;" + CRLF
			*! Close the CREATE TABLE command and switch to ALTER TABLE
			cCreateTable = cCreateTable + GDX_CMDMARGIN + GDX_INDENT + ")" + CRLF
			lUseAlterTable = .T.
		ENDIF
		IF lUseAlterTable
			cCreateTable = cCreateTable + GDX_CMDMARGIN + GDX_INDENT + "ALTER TABLE " + m.cTableName + " ADD COLUMN " + PADR(aAll_Fields(m.nInner_Loop, 1),MAX(LEN(aAll_Fields(m.nInner_Loop, 1)),GDX_FLDNAMEPAD)) + " " + aAll_Fields(m.nInner_Loop, 2)
		ELSE
			cCreateTable = cCreateTable + GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT + PADR(aAll_Fields(m.nInner_Loop, 1),MAX(LEN(aAll_Fields(m.nInner_Loop, 1)),GDX_FLDNAMEPAD)) + " " + aAll_Fields(m.nInner_Loop, 2)
		ENDIF
		DO CASE
			CASE aAll_Fields(m.nInner_Loop, 2) == "C"
				cCreateTable = m.cCreateTable + "(" + ;
								ALLTRIM(STR(aAll_Fields(m.nInner_Loop, 3))) + ")"
				IF aAll_Fields(m.nInner_Loop, 6)
					cCreateTable = m.cCreateTable + " NOCPTRANS"
				ENDIF
			CASE aAll_Fields(m.nInner_Loop, 2) == "M"
				IF aAll_Fields(m.nInner_Loop, 6)
					cCreateTable = m.cCreateTable + " NOCPTRANS"
				ENDIF
			CASE aAll_Fields(m.nInner_Loop, 2) == "N" OR ;
				aAll_Fields(m.nInner_Loop, 2) == "F"
				cCreateTable = m.cCreateTable + "(" + ;
				ALLTRIM(STR(aAll_Fields(m.nInner_Loop, 3))) + ;
				", " + ALLTRIM(STR(aAll_Fields(m.nInner_Loop, 4))) + ")"
			CASE aAll_Fields(m.nInner_Loop, 2) == "B"
				cCreateTable = m.cCreateTable + "(" + ;
				ALLTRIM(STR(aAll_Fields(m.nInner_Loop, 4))) + ")"
		ENDCASE

		IF aAll_Fields(m.nInner_Loop, 5)
			cCreateTable = m.cCreateTable + " NULL"
		ELSE
			cCreateTable = m.cCreateTable + " NOT NULL"
		ENDIF

		#IF CREATE_TABLE_SET_PROPS
			*! Get properties for fields
			IF !EMPTY(aAll_Fields[m.nInner_Loop,7])
				cCreateTable = m.cCreateTable + " CHECK " + aAll_Fields[m.nInner_Loop,7]
				*! SEA: Only insert the ERROR message if we have a CHECK clause
				IF !EMPTY(aAll_Fields[m.nInner_Loop,8])
					cCreateTable = m.cCreateTable + " ERROR " + aAll_Fields[m.nInner_Loop,8]
				ENDIF
			ENDIF

			IF !EMPTY(aAll_Fields(m.nInner_Loop, 9))
				cCreateTable = m.cCreateTable + " DEFAULT " + aAll_Fields[m.nInner_Loop,9]
			ENDIF
		#ENDIF

		IF m.nInner_Loop <> m.nNumberOfFields
			IF !lUseAlterTable
				cCreateTable = m.cCreateTable + m.cFieldLineEnd
			ELSE
				cCreateTable = cCreateTable + CRLF
			ENDIF
		ELSE
			IF !lUseAlterTable
				cCreateTable = m.cCreateTable + " ; &"+ "& " + CRLF + ;
					GDX_CMDMARGIN + GDX_INDENT + ")"
			ENDIF
		ENDIF
	ENDFOR
	*! Output the code to create the table
	=WriteFile(m.hOutFile, m.cCreateTable)

	IF TAGCOUNT() > 0
		*! Get Index Information
		#IF MAKE_INDEX_PASS
			*-- Start pass 3, create the table's indexes in a separate pass
			=WriteFile(m.hOutFile, CRLF + GDX_CMDMARGIN +                    "CASE tnPassNumber = 3")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "IF !USED('" + m.cTableName + "')")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + GDX_INDENT +       "USE " + m.cTableName + " AGAIN IN 0 ALIAS " + m.cTableName)
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "ENDIF")
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT +                 "SELECT " + m.cTableName)
			=WriteFile(m.hOutFile, "")
		#ENDIF

		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "***** " + BEGIN_INDEX_LOC + m.cTableName + " *****")
		cCollate = ""
		FOR nInner_Loop = 1 TO TAGCOUNT()
			cTag = UPPER(ALLTRIM(TAG(m.nInner_Loop)))
			IF m.cCollate <> IDXCOLLATE(m.nInner_Loop)
				cCollate = IDXCOLLATE(m.nInner_Loop)
				=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "SET COLLATE TO '" + m.cCollate + "'")
			ENDIF
			IF !EMPTY(m.cTag)
				DO CASE
					CASE PRIMARY(m.nInner_Loop)
						IF !EMPTY(SYS(2021, m.nInner_Loop))
							IF EMPTY(m.g_cFilterExp)
								=MessageBox(NOT_SUPPORTED_LOC, 64, WARNING_TITLE_LOC)
							ENDIF
							g_cFilterExp = GDX_CMDMARGIN + GDX_INDENT + m.g_cFilterExp + CRLF + ;
							               GDX_CMDMARGIN + GDX_INDENT + TABLE_NAME_LOC + m.cTableName + CRLF + ;
							               GDX_CMDMARGIN + GDX_INDENT + PRIMARY_KEY_LOC + SYS(14, m.nInner_Loop) + CRLF + ;
							               GDX_CMDMARGIN + GDX_INDENT + FILTER_EXP_LOC + SYS(2021, m.nInner_Loop)
						ENDIF
						=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "ALTER TABLE '" + m.cTableName + ;
							"' ADD PRIMARY KEY " + SYS(14, m.nInner_Loop) ;
							+ " TAG " + m.cTag + IIF(ALTER_TABLE_NOVALIDATE, " NOVALIDATE", ""))

					CASE CANDIDATE(m.nInner_Loop)
						IF EMPTY(SYS(2021, m.nInner_Loop))
							=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "INDEX ON " + SYS(14, m.nInner_Loop) + ;
								" TAG " + m.cTag + " CANDIDATE")
						ELSE
							=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "INDEX ON " + SYS(14, m.nInner_Loop) + ;
								" TAG " + m.cTag + " FOR " + SYS(2021, m.nInner_Loop) + ;
								+ " CANDIDATE")
						ENDIF
	
					CASE UNIQUE(m.nInner_Loop)
						IF(EMPTY(SYS(2021, m.nInner_Loop)))
							=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "INDEX ON " + SYS(14, m.nInner_Loop) + ;
								" TAG " + m.cTag + " UNIQUE")
						ELSE
							=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "INDEX ON " + SYS(14, m.nInner_Loop);
								+ " TAG " + m.cTag + " FOR " + SYS(2021, m.nInner_Loop) ;
								+ " UNIQUE")
						ENDIF
	
					OTHERWISE
						IF(EMPTY(SYS(2021, m.nInner_Loop)))
							=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "INDEX ON " + SYS(14, m.nInner_Loop) + ;
								" TAG " + m.cTag)
						ELSE
							=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "INDEX ON " + SYS(14, m.nInner_Loop);
								+ " TAG " + m.cTag + " FOR " + SYS(2021, m.nInner_Loop))
						ENDIF
				ENDCASE
			ELSE
				EXIT FOR
			ENDIF
		ENDFOR
	ENDIF
	
	IF lUseAlterTable
		*! ALTER TABLE commands can leave numerous deleted rows in the
		*! database container, so we should pack the dbc.  Note that the dbc
		*! must have been opened exclusively in order for this to work.
		=WriteFile(m.hOutFile, "")
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "***** " + PACK_DATABASE_LOC + " *****")
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "CLOSE TABLES")
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "PACK DATABASE")
	ENDIF

	*-- Start pass 2, create the table's properties
	=WriteFile(m.hOutFile, CRLF + GDX_CMDMARGIN + "CASE tnPassNumber = 2")
	
	*! Get Properties For Table
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "***** " + BEGIN_PROP_LOC + m.cTableName + " *****")
	FOR m.nInner_Loop = 1 TO m.nNumberOfFields
		cFieldAlias = aAll_Fields(m.nInner_Loop, 1)
		cFieldHeaderAlias = GDX_CMDMARGIN + GDX_INDENT + "ALTER TABLE '" + m.cTableName + "' ALTER COLUMN " + m.cFieldAlias

		#IF !CREATE_TABLE_SET_PROPS
			*! Get properties for fields
			IF !EMPTY(aAll_Fields[m.nInner_Loop,7])
				cTemp = m.cFieldHeaderAlias + " SET CHECK " + aAll_Fields[m.nInner_Loop,7]
				*! SEA: Only insert the ERROR message if we have a CHECK clause
				IF !EMPTY(aAll_Fields[m.nInner_Loop,8])
					cTemp = m.cTemp + " ERROR " + aAll_Fields[m.nInner_Loop,8]
				ENDIF
				cTemp = m.cTemp + " NOVALIDATE"
				=WriteFile(hOutFile, m.cTemp)
			ENDIF

			IF !EMPTY(aAll_Fields(m.nInner_Loop, 9))
				=WriteFile(hOutFile, m.cFieldHeaderAlias + " SET DEFAULT " + aAll_Fields[m.nInner_Loop,9] + " NOVALIDATE")
			ENDIF
		#ENDIF

		cFieldAlias = m.cTableName + "." + aAll_Fields(m.nInner_Loop, 1)
		*!*	cFieldHeaderAlias = GDX_CMDMARGIN + GDX_INDENT + [=DBSetProp('] + m.cFieldAlias + [', 'Field', ]
		cFieldHeaderAlias = GDX_CMDMARGIN + GDX_INDENT + [DBSetProp('] + m.cFieldAlias + [', 'Field', ]
		cTemp = DBGETPROP(m.cFieldAlias, "Field", "Caption")
		IF !EMPTY(cTemp)
			cTemp = STRTRAN(m.cTemp, ["], ['])
			=WriteFile(hOutFile, m.cFieldHeaderAlias + ['Caption', "] + m.cTemp + [")])
		ENDIF
		cTemp = DBGETPROP(m.cFieldAlias, "Field", "Comment")
		IF !EMPTY(m.cTemp)
			cTemp = STRTRAN(m.cTemp, ["], ['])
			*! Strip Line Feeds
			cTemp = STRTRAN(m.cTemp, CHR(10)) 
			*! Convert Carriage Returns To Programmatic Carriage Returns
			cTemp = STRTRAN(m.cTemp, CHR(13), '" + CHR(13) + "')
			=WriteFile(m.hOutFile, m.cFieldHeaderAlias + ['Comment', "] + m.cTemp + [")])
		ENDIF

		*! The remaining field-level properties don't apply to version 3
*!*  		IF "VISUAL FOXPRO" $ upper(version(1)) and !"VISUAL FOXPRO 03" $ upper(version(1)) jme 18/01/2009
			* We're not running version 3, so we must be running 5 or higher
			cTemp = DBGETPROP(m.cFieldAlias, "Field", "InputMask")
			IF !EMPTY(m.cTemp)
				m.cTemp = STRTRAN(m.cTemp, ["], ['])
				=WriteFile(m.hOutFile, m.cFieldHeaderAlias + ['InputMask', "] + m.cTemp + [")])
			ENDIF
			cTemp = DBGETPROP(m.cFieldAlias, "Field", "Format")
			IF !EMPTY(m.cTemp)
				cTemp = STRTRAN(m.cTemp, ["], ['])
				=WriteFile(m.hOutFile, m.cFieldHeaderAlias + ['Format', "] + m.cTemp + [")])
			ENDIF
			cTemp = DBGETPROP(m.cFieldAlias, "Field", "DisplayClass")
			IF !EMPTY(m.cTemp)
				cTemp = STRTRAN(m.cTemp, ["], ['])
				=WriteFile(m.hOutFile, m.cFieldHeaderAlias + ['DisplayClass', "] + m.cTemp + [")])
			ENDIF
			cTemp = DBGETPROP(m.cFieldAlias, "Field", "DisplayClassLibrary")
			IF !EMPTY(m.cTemp)
				cTemp = STRTRAN(m.cTemp, ["], ['])
				=WriteFile(m.hOutFile, m.cFieldHeaderAlias + ['DisplayClassLibrary', "] + m.cTemp + [")])
			ENDIF
*!*  		ENDIF
	ENDFOR
	
	cTemp = DBGETPROP(m.cTableName, "Table", "Comment")
	IF !EMPTY(m.cTemp)
		cTemp = STRTRAN(m.cTemp, ["], ['])
		*! Strip Line Feeds
		cTemp = STRTRAN(m.cTemp, CHR(10))
		*! Convert Carriage Returns To Programmatic Carriage Returns
		cTemp = STRTRAN(m.cTemp, CHR(13), '" + CHR(13) + "')
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + [DBSetProp('] + m.cTableName + [', 'Table', ] + ['Comment', "] + m.cTemp + [")])
	ENDIF

	cTemp = DBGETPROP(m.cTableName, "Table", "DeleteTrigger")
	IF !EMPTY(m.cTemp)
		=WriteFile(hOutFile, GDX_CMDMARGIN + GDX_INDENT + "CREATE TRIGGER ON '" + m.cTableName + ;
							  "' FOR DELETE AS " + m.cTemp)
	ENDIF

	cTemp = DBGETPROP(m.cTableName, "Table", "InsertTrigger")
	IF !EMPTY(m.cTemp)
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "CREATE TRIGGER ON '" + m.cTableName + ;
							  "' FOR INSERT AS " + m.cTemp)
	ENDIF

	cTemp = DBGETPROP(m.cTableName, "Table", "UpdateTrigger")
	IF !EMPTY(m.cTemp)
		=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "CREATE TRIGGER ON '" + m.cTableName + ;
							  "' FOR UPDATE AS " + m.cTemp)
	ENDIF

	cTemp = DBGETPROP(m.cTableName, "Table", "RuleExpression")
	IF !EMPTY(m.cTemp)
		cError = DBGETPROP(m.cTableName, "Table", "RuleText")
		IF !EMPTY(cError)
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "ALTER TABLE '" + m.cTableName + ;
								  "' SET CHECK " + m.cTemp + " ERROR " + ;
								  m.cError)
		ELSE
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + GDX_INDENT + "ALTER TABLE '" + m.cTableName + ;
								  "' SET CHECK " + m.cTemp)
		ENDIF
	ENDIF
	
	=WriteFile(m.hOutFile, GDX_CMDMARGIN + "ENDCASE")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")
	=WriteFile(m.hOutFile, "")

RETURN

**************************************************************************
**
** Function Name: GETVIEW(, )
** Creation Date: 1994.12.01
** Purpose        :
**
**              To take an existing FoxPro View, and generate an
**              output program that can be used to "re-create" that view.
**
** Parameters:
**
**      cViewName        A character string representing the name of the 
**                       existing view
**
**      hOutFile         The handle of the output file
**
**      cProcName        The name of the procedure within which we wrap the
**                       code that recreates the view
**
** Modification History:
**
**  1994.12.01  JHL      Created Program, runs on Build 329 of FoxPro 3.0
**  1994.12.02  KRT      Added to GenDBC, removed third parameter, cleaned up
**  1994.12.08  KRT      Assume Database is open to speed up operation
**  1996.04.12  KRT      Added new properties for Visual FoxPro 5.0
**                       Prepared / CompareMemo / FetchAsNeeded
**	 1996.05.14  KRT      Added more properties for views
**  1996.05.16  KRT      Adjusted for return a blank string instead of a logical
**                       value on Prepared, etc.. if the field does not exist
**                       in the database (Version 3.0 database converted to 5.0)
**  1996.05.16  KRT      Added the DataType property
**  1996.06.26  KRT      Added support for ParameterList
**  1996.07.19  KRT      Added support for Offline Views
**  1996.08.07  KRT      Added support for BatchUpdateCount, Comment
**  1997.02.24  SEA      Output code within a PROCEDURE / ENDPROC block to
**                       avoid error when compiling procedures > 64K
**                       database re-creation program a procedure
**  1997.02.24  SEA      Write directly to a single output file, the handle 
**                       of which is passed in hOutFile
**  1997.09.05  SEA      Removed duplicate "FetchSize" output line (bug fix).
**  1997.09.05  SEA      Added check for VFP 3.0 before attempting to retrieve
**                       5.0-specific properties (bug fix).  (Thanks to
**                       Nancy Folsom for catching this!)
**  1997.09.05  SEA      Added code to handle view names that contain spaces when
**                       creating the view-recreation procedure name
**  1997.12.07  SEA      Added code to initialize local variables for each parameter
**                       in a parameterized view.  Note that this only works if you
**                       have added each parameter to the view's parameters list,
**                       either through the View Designer (Query/View Parameters) or
**                       with DBSetProp().
**  jme 18/01/2009 personnalisation : suppression de commandes inutiles si la vue 
**                 n'est pas updatable
***************************************************************************************
PROCEDURE GetView
	LPARAMETERS cViewName, hOutFile, cProcName, nindice, tllaissenombdd

	PRIVATE ALL EXCEPT g_*


*DIMENSION g_ltsqlvue( m.nTotal_Views) && jme 09/02/2009
*g_ltsqlvue = ""
*WAIT WINDOW "nloop =" + STR(m.nindice) TIMEOUT 3
*tllaissenombdd  jme 19/02/2009 .T. si on laisse le nom d ela bdd dans la clause FROM
*                               par dfaut NON

	*! Get View Information for later use
	nSourceType       = DBGetProp(m.cViewName, 'View', 'SourceType')
	cConnectName      = ALLTRIM(DBGetProp(m.cViewName, 'View', 'ConnectName'))
	cSQL              = ALLTRIM(DBGetProp(m.cViewName, 'View', 'SQL'))
	cnUpdateType      = ALLTRIM(STR(DBGetProp(m.cViewName, 'View', 'UpdateType')))
	cnWhereType       = ALLTRIM(STR(DBGetProp(m.cViewName, 'View', 'WhereType')))
	clFetchMemo       = IIF(DBGetProp(m.cViewName, 'View', 'Fetchmemo'),'.T.','.F.')
	clShareConnection = IIF(DBGetProp(m.cViewName, 'View', 'ShareConnection'),'.T.','.F.')
	clSendUpdates     = IIF(DBGetProp(m.cViewName, 'View', 'SendUpdates'),'.T.','.F.')
	cnUseMemoSize     = ALLTRIM(STR(DBGetProp(m.cViewName, 'View', 'UseMemoSize')))
	cnFetchSize       = ALLTRIM(STR(DBGetProp(m.cViewName, 'View', 'FetchSize')))
	cnMaxRecords      = ALLTRIM(STR(DBGetProp(m.cViewName, 'View', 'MaxRecords')))
	ccTables          = ALLTRIM(DBGetProp(m.cViewName, 'View', 'Tables'))
	nnbtable = 1 + OCCURS(",", m.ccTables) && jme 18/01/2009
  nnposwhere = 0 && position du WHERE dans la chaine Csql
  IF m.nnbtable = 1
    * on enlve le nom de la table pour chaque champ dans la commande SQL (il n'y a alors
    * pas d'ambiguit). On enlve aussi le nom de la base de donnes dans la clause FROM
    * et ce brave garon remet le nom de la table en tant qu'alias local !!
    ndebuttable = 8
    nfintable = AT(".", m.cSQL)
    nlongueur = m.nfintable - m.ndebuttable
    ccaccu = SUBSTR(m.cSQL, m.ndebuttable, m.nlongueur)
    nlaccu = AT("!", m.ccaccu)
    IF m.nlaccu = 0
      cnombdd = ""
      cnomtable = m.ccaccu
      * on peut aussi avoir le nom de la bdd dans la clause FROM voir plus loin
    ELSE && m.nlaccu = 0
      cnombdd = LEFT(m.ccaccu, m.nlaccu -1)
      cnomtable = SUBSTR( m.ccaccu, m.nlaccu+1)
    ENDIF && m.nlaccu = 0
    IF !EMPTY(m.cnomtable)
      cSQL = STRTRAN( m.cSQL, m.cnomtable+".", "", -1, -1, 1)
      cSQL = STRTRAN( m.cSQL, " "+m.cnomtable, "", -1, -1, 1)
    ENDIF && !EMPTY(m.cnomtable)
    IF !EMPTY( m.cnombdd)
      cSQL = STRTRAN( m.cSQL, m.cnombdd+"!", "", -1, -1, 1)
    ENDIF && !EMPTY( m.cnombdd)
    cSQL = STRTRAN(m.cSQL, ",", ", ")
    cSQL = STRTRAN(m.cSQL, ",  ", ", ")
  ENDIF && m.nnbtable = 1
  * on essaie de formater la ligne
  STORE 1 TO nlpos
  STORE 1 TO nlnbindent
  nllongueur = LEN(m.cSQL)
  DO WHILE m.nlpos < m.nllongueur -5
    nlposcar = m.nlpos
    nlposmax = m.nlposcar + 80
    nlposmin = m.nlposcar + 8
    DO WHILE m.nlposcar < m.nlposmax AND SUBSTR(m.cSQL, m.nlposcar, 7) <> " WHERE " ;
     AND SUBSTR(m.cSQL, m.nlposcar, 6) <> " FROM " AND ;
      SUBSTR(m.cSQL, m.nlposcar, 7) <> " ORDER " AND SUBSTR(m.cSQL, m.nlposcar, 7) <> " GROUP "
      nlposcar = m.nlposcar + 1
    ENDDO && m.nlindice < 80 AND SUBSTR(m.cSQL, m.nlindice, 7) <> " WHERE "
    DO CASE 
    CASE SUBSTR(m.cSQL, m.nlposcar, 7) == " WHERE "
      * on revient  la ligne avant WHERE
      nlnbindent = 1
      cSQL = LEFT(m.cSQL, m.nlposcar)+ " ;" + CRLF +GDX_CMDMARGIN + REPLICATE(GDX_INDENT, m.nlnbindent)+ ;
       "WHERE"+ SUBSTR(m.cSQL, m.nlposcar + 6)
      nnposwhere = m.nlposcar + 6 && on veut supprimer les () plus loin
      nlpos = m.nlposcar + 9
    CASE SUBSTR(m.cSQL, m.nlposcar, 6) == " FROM "
      * on revient  la ligne avant FROM
      * on peut trouver ici le nom de la bdd; Le troisime paramtre d'appel de gendbcxm
      * indique si on l'enlever (le dfaut) ou non
      * tllaissenombdd si .T. on laisse le nom de la bdd dans les
      * clauses FROM des vues. .F. par dfaut : cela permet de
      * gnrer des vues qui seront stockes dans une autre bbd 
      * que celle utilise par gendbcxm (si on veut ajouter une
      * on la met dans une BDD 'temporaire', on gnre le programme
      * et on execute le programme gnr avec la bdd correcte.
      nlposreprise = m.nlposcar + 6
      IF m.tllaissenombdd = .F. AND EMPTY( m.cnombdd)
        * on ne veut pas que le nom de la bdd apparaisse dans la clause FROM
        * et on ne l'a pas encore trouv
        nlaccu = AT("!", SUBSTR(m.cSQL, m.nlposreprise))
        IF m.nlaccu > 0
          * le ! ne peut tre que dans la clause FROM
          nlposreprise = m.nlposreprise + m.nlaccu
        ENDIF && m.nlaccu > 0
      ENDIF && m.tllaissenombdd = .F.
      nlnbindent = 1
      cSQL = LEFT(m.cSQL, m.nlposcar)+ " ;" + CRLF +GDX_CMDMARGIN + REPLICATE(GDX_INDENT, m.nlnbindent)+ ;
       "FROM "+ SUBSTR(m.cSQL, m.nlposreprise)
*!*        nlpos = m.nlposcar + 8
      nlpos = m.nlposreprise + 3
    CASE SUBSTR(m.cSQL, m.nlposcar, 7) == " ORDER "
      * on revient  la ligne avant ORDER BY
      nlnbindent = 1
      cSQL = LEFT(m.cSQL, m.nlposcar)+ " ;" + CRLF +GDX_CMDMARGIN + REPLICATE(GDX_INDENT, m.nlnbindent)+ ;
       "ORDER"+ SUBSTR(m.cSQL, m.nlposcar + 6) && 1 seul espace entre order et by
      nlpos = m.nlposcar + 8
    CASE SUBSTR(m.cSQL, m.nlposcar, 7) == " GROUP "
      * on revient  la ligne avant GROUP BY
      nlnbindent = 1
      cSQL = LEFT(m.cSQL, m.nlposcar)+ " ;" + CRLF +GDX_CMDMARGIN + REPLICATE(GDX_INDENT, m.nlnbindent)+ ;
       "GROUP"+ SUBSTR(m.cSQL, m.nlposcar + 6)
      nlpos = m.nlposcar + 8
*!*      CASE 
*!*        
    OTHERWISE
      * on limite les lignes aux environs de 80 caractres d'abord en revenant en arrire
      * puis en cas de problme en repartant vers l'avant
      nlnbindent = MIN(m.nlnbindent + 1, 8)
      nlposaccu = m.nlposcar
      DO WHILE m.nlposcar > m.nlposmin AND SUBSTR(m.cSQL, m.nlposcar,1) <> " "
        nlposcar = m.nlposcar - 1
      ENDDO && m.nlposcar > 8 AND SUBSTR(m.cSQL, m.nlposcar,1) <> " "
      IF SUBSTR(m.cSQL, m.nlposcar,1) == " "
        cSQL = LEFT(m.cSQL, m.nlposcar)+ ";" + CRLF + GDX_CMDMARGIN + REPLICATE(GDX_INDENT, m.nlnbindent)+ ;
         SUBSTR(m.cSQL, m.nlposcar + 1)
      ELSE && SUBSTR(m.cSQL, m.nlposcar,1) == " "
        nlposcar = m.nlposaccu
        DO WHILE m.nlposcar < m.nllongueur -5 AND SUBSTR(m.cSQL, m.nlposcar,1) <> " "
          nlposcar = m.nlposcar + 1
        ENDDO && m.nlposcar < m.nllongueur -5 AND SUBSTR(m.cSQL, m.nlposcar,1) <> " "
        IF SUBSTR(m.cSQL, m.nlposcar,1) == " "
          cSQL = LEFT(m.cSQL, m.nlposcar)+ ";" + CRLF + GDX_CMDMARGIN + REPLICATE(GDX_INDENT, m.nlnbindent)+ ;
           SUBSTR(m.cSQL, m.nlposcar + 1)
        ENDIF && SUBSTR(m.cSQL, m.nlposcar,1) == " "
      ENDIF && SUBSTR(m.cSQL, m.nlposcar,1) == " "
      nlpos = m.nlposcar + 5 + m.nlnbindent
    ENDCASE
  ENDDO && m.nlpos < m.nllongueur
  * jme 01/02/2009 on essaie d'enlever les () autour de la clause WHERE
  STORE m.nnposwhere + 8 TO nnposwhere, nlposcar
  m.nllongueur = LEN(m.cSQL) && la longueur de cSQL a surement chang !
  DO WHILE m.nlposcar <= m.nllongueur AND SUBSTR(m.cSQL, m.nlposcar, 7) <> " ORDER " ;
   SUBSTR(m.cSQL, m.nlposcar, 7) <> " GROUP "
    nlposcar = m.nlposcar + 1
  ENDDO && m.nlpos < m.nllongueur -5
  nllongu = m.nlposcar - m.nnposwhere
  ccaccu = STRTRAN(SUBSTR(m.cSQL, m.nnposwhere, m.nllongu), "(", "", 1)
  nlpos = LEN(m.ccaccu)
  DO WHILE m.nlpos>1 AND SUBSTR(m.ccaccu, m.nlpos, 1) <> ")"
    nlpos = m.nlpos - 1
  ENDDO && m.nlpos>1
  IF m.nlpos>1
    ccaccu = LEFT(m.ccaccu, m.nlpos -1)+ SUBSTR(m.ccaccu, m.nlpos+1)
  ENDIF && m.nlpos>1
*  WAIT WINDOW m.ccaccu
*!*    cSQL = STRTRAN(LEFT(m.cSQL, m.nnposwhere)+ m.ccaccu + ;
*!*     IIF(m.nlposcar < m.nllongueur, SUBSTR(m.cSQL, m.nlposcar), ""), "  ", " ")
  cSQL = LEFT(m.cSQL, m.nnposwhere)+ m.ccaccu + ;
   IIF(m.nlposcar < m.nllongueur, SUBSTR(m.cSQL, m.nlposcar), "")
  cSQL = STRTRAN(STRTRAN(m.cSQL, "   ", "  "), "  ", " ")
  g_ltsqlvue( m.nindice) = cSQL && jme 09/02/2009

  cParams           = ALLTRIM(DBGetProp(m.cViewName, 'View', 'ParameterList'))
	cComment          = DBGETPROP(m.cViewName, 'View', 'Comment')
	IF !EMPTY(m.cComment )
		cComment = STRTRAN(m.cComment , ["], ['])
		*! Strip Line Feeds
		cComment = STRTRAN(m.cComment , CHR(10)) 
		*! Convert Carriage Returns To Programmatic Carriage Returns
		cComment = STRTRAN(m.cComment , CHR(13), '" + CHR(13) + "')
	ENDIF
	cnBatchUpdateCount = ALLTRIM(STR(DBGetProp(m.cViewName, 'View', 'BatchUpdateCount')))

	*! Generate Heading
	=WriteFile(m.hOutFile, "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "**************************************************")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "** " + BEGIN_VIEW_LOC + m.cViewName)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "**************************************************")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.cProcName)
*  =WriteFile(m.hOutFile, GDX_CMDMARGIN + "LOCAL lcvue") jme 01/02/2009

	IF !EMPTY(m.cParams)
		*! Generate code to create parameters before we create the view, so that VFP
		*! doesn't ask the user for the parameters' values.
		cParamList = cParams + "; "
		cLocalsStr = GDX_CMDMARGIN + "LOCAL "
		cParamsStr = ""
		DO WHILE !EMPTY(cParamList)
			cParamName = left(cParamList,at(",",cParamList)-1)
			cDataType  = substr(cParamList,at(",",cParamList)+2,1)
			cCrtParam  = cParamName + " = " + DataValue(cDataType)
			cLocalsStr = cLocalsStr + cParamName + ","
			cParamsStr = cParamsStr + GDX_CMDMARGIN + cCrtParam + CRLF
			cParamList = substr(cParamList,at(";",cParamList)+1)
		ENDDO
		*! Strip the trailing comma from the LOCAL statement
		cLocalsStr = left(cLocalsStr,len(cLocalsStr)-1)
		*! Write the code
		=WriteFile(m.hOutFile, cLocalsStr)
		=WriteFile(m.hOutFile, cParamsStr)
	ELSE
    =WriteFile(m.hOutFile, "") && saut de ligne
  ENDIF

	*! Generate CREATE VIEW command
  =WriteFile(m.hOutFile, GDX_CMDMARGIN + 'lcvue = "'+ ALLTRIM(m.cViewName) +'"')
*!*  	cCreateString = GDX_CMDMARGIN + 'CREATE SQL VIEW "' + ALLTRIM(m.cViewName) + '" ; ' + CRLF
  cCreateString = GDX_CMDMARGIN + 'CREATE SQL VIEW (m.lcvue) ; ' + CRLF
	IF m.nSourceType != 1     && If it isn't a local view
		cCreateString = m.cCreateString + GDX_CMDMARGIN + GDX_INDENT + 'REMOTE '
		IF !EMPTY(m.cConnectName)
			cCreateString = m.cCreateString + 'CONNECTION "' + m.cConnectName + '" ; '+CRLF
		ENDIF
	ENDIF
	cCreateString = m.cCreateString + GDX_CMDMARGIN + GDX_INDENT + 'AS '+ m.cSQL + CRLF

	=WriteFile(m.hOutFile, m.cCreateString)
*!*    =WriteFile(m.hOutFile, GDX_CMDMARGIN +"lcv = '" + m.cViewName + "'") && jme 29/05/2008 jme 01/02/2009
*!*    =WriteFile(m.hOutFile, GDX_CMDMARGIN +"lcp = 'View'") && jme 29/05/2008

	*! GENERATE code to Set View Level Properties
*!*  	cViewDBSetPrefix = GDX_CMDMARGIN + [DBSetProp('] + m.cViewName + [', 'View', ]
*!*    cViewDBSetPrefix = GDX_CMDMARGIN + [DBSetProp( m.lcv, m.lcp, ]
  cViewDBSetPrefix = GDX_CMDMARGIN + [DBSetProp( m.lcvue, m.lcv, ]
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['UpdateType', ] + m.cnUpdateType + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['WhereType', ] + m.cnWhereType + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['FetchMemo', ] + m.clFetchMemo + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['SendUpdates', ] + m.clSendUpdates + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['UseMemoSize', ] + m.cnUseMemoSize + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['FetchSize', ] + m.cnFetchSize + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['MaxRecords', ] + m.cnMaxRecords + [)])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['Tables', '] + m.ccTables + [')])
*!*  	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['FetchSize', ] + m.cnFetchSize + [)])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvudt, ] + m.cnUpdateType + [)  &]+[& UpdateType])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvwt, ] + m.cnWhereType + [)  &]+[& WhereType])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvfm, ] + m.clFetchMemo + [)  &]+[& FetchMemo])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvsu, ] + m.clSendUpdates + [)  &]+[& SendUpdates])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvums, ] + m.cnUseMemoSize + [)  &]+[& UseMemoSize])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvfs, ] + m.cnFetchSize + [)  &]+[& FetchSize])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvmr, ] + m.cnMaxRecords + [)  &]+[& MaxRecords])
  =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvt, '] + m.ccTables + [')  &]+[& Tables])
	IF !EMPTY(m.cParams)
*!*  		=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['ParameterList', "] + m.cParams + [")])
    =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvpl, "] + m.cParams + [")  &]+[& ParameterList])
	ENDIF
	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvcom, "] +  m.cComment + [")  &]+[& Comment])
	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvbuc, ] + m.cnBatchUpdateCount + [)  &]+[& BatchUpdateCount])
	=WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvsc, ] + m.clShareConnection + [)  &]+[& ShareConnection])

	*! The remaining view properties don't apply to version 3
*!*  	IF "VISUAL FOXPRO" $ upper(version(1)) and !"VISUAL FOXPRO 03" $ upper(version(1))
		clPrepared = IIF(!EMPTY(DBGetProp(m.cViewName, 'View', 'Prepared')), '.T.', '.F.')
		clCompareMemo = IIF(!EMPTY(DBGetProp(m.cViewName, 'View', 'CompareMemo')), '.T.', '.F.')
		clFetchAsNeeded = IIF(!EMPTY(DBGetProp(m.cViewName, 'View', 'FetchAsNeeded')), '.T.', '.F.')
		lOffline = DBGetProp(m.cViewName, 'View', 'Offline')
*!*  		=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['Prepared', ] + m.clPrepared + [)])
*!*  		=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['CompareMemo', ] + m.clCompareMemo + [)])
*!*  		=WriteFile(m.hOutFile, m.cViewDBSetPrefix + ['FetchAsNeeded', ] + m.clFetchAsNeeded + [)])
    =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvp, ] + m.clPrepared + [)  &]+[& Prepared])
    =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvcm, ] + m.clCompareMemo + [)  &]+[& CompareMemo])
    =WriteFile(m.hOutFile, m.cViewDBSetPrefix + [m.lcvfan, ] + m.clFetchAsNeeded + [)  &]+[& FetchAsNeeded])
		IF m.lOffline
			=WriteFile(m.hOutFile, GDX_CMDMARGIN + 'CREATEOFFLINE("' + m.cViewName + '")')
		ENDIF
*!*  	ENDIF

  IF m.clSendUpdates == '.T.' && jme 18/01/2009
	*! GENERATE code to Set Field Level Properties
  *! jme 18/01/2009 seulement si la vue est updatable, sinon le paramtrage des champs
  *!                 ne sert  rien
  	USE (DBC()) AGAIN IN 0 ALIAS GenViewCursor EXCLUSIVE
  	SELECT GenViewCursor
  	LOCATE FOR ALLTRIM(UPPER(GenViewCursor.ObjectName)) == m.cViewName AND ;
      	GenViewCursor.ObjectType = 'View'
  	nObjectId = GenViewCursor.ObjectId
  	SELECT ObjectName FROM GenViewCursor ;
    	WHERE GenViewCursor.ParentId = m.nObjectId ;
  			INTO ARRAY aViewFields
  	USE in GenViewCursor
  	=WriteFile(m.hOutFile, "")
*!*      =WriteFile(m.hOutFile, GDX_CMDMARGIN +"lcp = 'Field'") && jme 29/05/2008 jme 01/02/2009
*!*    	=WriteFile(m.hOutFile, GDX_CMDMARGIN + '*!* Field Level Properties for ' + m.cViewName) jme 22/01/2009

  	IF _TALLY # 0
  		FOR m.nLoop = 1 TO ALEN(aViewFields, 1)
  			cFieldAlias = m.cViewName + "." + ALLTRIM(aViewFields(nLoop, 1))
  			clKeyField = IIF(DBGetProp(m.cFieldAlias, 'Field', 'KeyField'),'.T.','.F.')
  			clUpdatable = IIF(DBGetProp(m.cFieldAlias, 'Field', 'Updatable'),'.T.','.F.')
  			ccUpdateName = ALLTRIM(DBGetProp(m.cFieldAlias, 'Field', 'UpdateName'))
  *!*  			cViewFieldSetPrefix = GDX_CMDMARGIN + [DBSetProp('] + m.cFieldAlias + [', 'Field', ]
*!*          cViewFieldSetPrefix = GDX_CMDMARGIN + [DBSetProp(m.lcf, m.lcp, ] && jme 29/05/2008 jme 01/02/2009
*!*          =WriteFile(m.hOutFile, GDX_CMDMARGIN +"lcf = '"+ m.cFieldAlias+ "'") && jme 29/05/2008 jme 01/02/2009
        cViewFieldSetPrefix = GDX_CMDMARGIN + [DBSetProp(m.lcp, m.lcf, ] && jme 01/02/2009
        =WriteFile(m.hOutFile, GDX_CMDMARGIN +"lcp = '"+ m.cFieldAlias+ "'") && jme 01/02/2009

*!*    			=WriteFile(m.hOutFile, GDX_CMDMARGIN + '* Props for the '+m.cFieldAlias+' field.') jme 01/02/2009
  *!*  			=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['KeyField', ] + m.clKeyField + [)])
  *!*  			=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['Updatable', ] + m.clUpdatable + [)])
  *!*  			=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['UpdateName', '] + m.ccUpdateName + [')])
        =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lckf, ] + m.clKeyField + [)  &]+[& KeyField]) && jme 29/05/2008
        =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lcupd, ] + m.clUpdatable + [)  &]+[& Updatable])
        =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lcupn, '] + m.ccUpdateName + [')  &]+[& UpdateName])
  			cTemp = DBGETPROP(m.cFieldAlias, "Field", "RuleExpression")
  			IF !EMPTY(m.cTemp)
  				cTemp = STRTRAN(m.cTemp, ["], ['])
  *!*  				=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['RuleExpression', "]+m.cTemp+[")])
          =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lcrle, "]+m.cTemp+[")  &]+[& RuleExpression])
  			ENDIF
  			
  			cTemp = DBGETPROP(m.cFieldAlias, "Field", "RuleText")
  			IF !EMPTY(m.cTemp)
  				cTemp = STRTRAN(m.cTemp, ["], ['])
  *!*  				=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['RuleText', "]+m.cTemp+[")])
          =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lcrlt, "]+m.cTemp+[") &]+[& RuleText])
  			ENDIF
  			*-- SAS
  			* cTemp = DBGETPROP(m.cFieldAlias, "Field", "Caption")
  			cTemp = ALLTRIM(DBGETPROP(m.cFieldAlias, "Field", "Caption"))
  			*-- SAS
  			IF !EMPTY(m.cTemp)
  				cTemp = STRTRAN(m.cTemp, ["], ['])
  *!*  				=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['Caption', "] + m.cTemp + [")])
          =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lccap, "] + m.cTemp + [")  &]+[& Caption])
  			ENDIF
  			cTemp = DBGETPROP(m.cFieldAlias, "Field", "Comment")
  			IF !EMPTY(m.cTemp)
  				cTemp = STRTRAN(m.cTemp, ["], ['])
  				*! Strip Line Feeds
  				cTemp = STRTRAN(m.cTemp, CHR(10)) 
  				*! Convert Carriage Returns To Programmatic Carriage Returns
  				cTemp = STRTRAN(m.cTemp, CHR(13), '" + CHR(13) + "')
  *!*  				=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['Comment', "] + m.cTemp + [")])
          =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lccom, "] + m.cTemp + [")  &]+[& Comment])
  			ENDIF
  			cTemp = DBGETPROP(m.cFieldAlias, "Field", "DataType")
  			IF !EMPTY(m.cTemp)
  *!*  				=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['DataType', "] + m.cTemp + [")])
          =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lcdt, "] + m.cTemp + [")  &]+[& DataType])
  			ENDIF
  			cTemp = DBGETPROP(m.cFieldAlias, "Field", "DefaultValue")
  			IF !EMPTY(m.cTemp)
  				cTemp = STRTRAN(m.cTemp, ["], ['])
  *!*  				=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['DefaultValue', "] + m.cTemp + [")])
          =WriteFile(m.hOutFile, m.cViewFieldSetPrefix + [m.lcdfv, "] + m.cTemp + [")  &]+[& DefaultValue])
  			ENDIF

  			*! The remaining field-level properties don't apply to version 3
*!*    			IF "VISUAL FOXPRO" $ upper(version(1)) and !"VISUAL FOXPRO 03" $ upper(version(1)) jme 18/01/2009
  				cTemp = DBGETPROP(m.cFieldAlias, "Field", "InputMask")
  				IF !EMPTY(m.cTemp)
  					cTemp = STRTRAN(m.cTemp, ["], ['])
  					=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['InputMask', "] + m.cTemp + [")])
  				ENDIF
  				cTemp = DBGETPROP(m.cFieldAlias, "Field", "Format")
  				IF !EMPTY(m.cTemp)
  					cTemp = STRTRAN(m.cTemp, ["], ['])
  					=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['Format', "] + m.cTemp + [")])
  				ENDIF
  				cTemp = DBGETPROP(m.cFieldAlias, "Field", "DisplayClass")
  				IF !EMPTY(m.cTemp)
  					cTemp = STRTRAN(m.cTemp, ["], ['])
  					=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['DisplayClass', "] + m.cTemp + [")])
  				ENDIF
  				cTemp = DBGETPROP(m.cFieldAlias, "Field", "DisplayClassLibrary")
  				IF !EMPTY(m.cTemp)
  					cTemp = STRTRAN(m.cTemp, ["], ['])
  					=WriteFile(m.hOutFile, m.cViewFieldSetPrefix + ['DisplayClassLibrary', "] + m.cTemp + [")])
  				ENDIF
*!*    			ENDIF
  		ENDFOR && m.nLoop = 1 TO ALEN(aViewFields, 1)
	  ENDIF && _TALLY # 0
  ELSE
    =WriteFile(m.hOutFile, GDX_PRCMARGIN + GDX_CMDMARGIN+ ;
     "* pas de proprits sur les champs : vue en lecture seule")
  ENDIF && clSendUpdates
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")

RETURN

**************************************************************************
**
** Function Name: GETCONN(, )
** Creation Date: 1995.01.03
** Purpose        :
**
**              To take an existing FoxPro Connection, and generate
**              an output program that can be used to "re-create" that connection.
**
** Parameters:
**
**      cConnectionName  A character string representing the name of the 
**                       existing connection
**
**      hOutFile         The handle of the output file
**
**      cProcName        The name of the procedure within which we wrap the
**                       code that recreates the connection
**
** Modification History:
**
**  1995.01.03  JHL      Created Program, runs on Build 329 of FoxPro 3.0
**  1995.01.05  KRT      Incorporated into GenDBC with modifications
**  1996.04.12  KRT      Added new property for Visual FoxPro 5.0 (Database)
**  1997.02.24  SEA      Output code within a PROCEDURE / ENDPROC block to
**                       avoid error when compiling procedures > 64K
**                       database re-creation program a procedure
**  1997.02.24  SEA      Write directly to a single output file, the handle 
**                       of which is passed in hOutFile
**  1997.09.05  SEA      When creating the procedure name, replace spaces
**                       with underscores.
***************************************************************************************
PROCEDURE GetConn
	LPARAMETERS cConnectionName, hOutFile, cProcName

	PRIVATE ALL EXCEPT g_*

	*! Get Connection Information for later use
	clAsynchronous = IIF(DBGetProp(m.cConnectionName, 'Connection', 'Asynchronous'),'.T.','.F.')
	clBatchMode = IIF(DBGetProp(m.cConnectionName, 'Connection', 'BatchMode'),'.T.','.F.')
	ccComment = ALLTRIM(DBGetProp(m.cConnectionName, 'Connection', 'Comment'))
	ccConnectString = ALLTRIM(DBGetProp(m.cConnectionName, 'Connection', 'ConnectString'))
	cnConnectTimeOut = ALLTRIM(STR(DBGetProp(m.cConnectionName, 'Connection', 'ConnectTimeOut')))
	ccDataSource = ALLTRIM(DBGetProp(m.cConnectionName, 'Connection', 'DataSource'))
	cnDispLogin = ALLTRIM(STR(DBGetProp(m.cConnectionName, 'Connection', 'DispLogin')))
	clDispWarnings = IIF(DBGetProp(m.cConnectionName, 'Connection', 'DispWarnings'),'.T.','.F.')
	cnIdleTimeOut = ALLTRIM(STR(DBGetProp(m.cConnectionName, 'Connection', 'IdleTimeOut')))
	ccPassword = ALLTRIM(DBGetProp(m.cConnectionName, 'Connection', 'Password'))
	cnQueryTimeOut = ALLTRIM(STR(DBGetProp(m.cConnectionName, 'Connection', 'QueryTimeOut')))
	cnTransactions = ALLTRIM(STR(DBGetProp(m.cConnectionName, 'Connection', 'Transactions')))
	ccUserId = ALLTRIM(DBGetProp(m.cConnectionName, 'Connection', 'UserId'))
	cnWaitTime = ALLTRIM(STR(DBGetProp(m.cConnectionName, 'Connection', 'WaitTime')))
	ccDatabase = DBGetProp(m.cConnectionName, 'Connection', 'Database')

	*! Generate Heading
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "**************************************************")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "** " + BEGIN_CONNECTIONS_LOC + " " + m.cConnectionName)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "**************************************************")
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "PROCEDURE " + m.cProcName)

	*! Generate CREATE Connection command
	cCreateString = GDX_CMDMARGIN + 'CREATE CONNECTION '+ALLTRIM(m.cConnectionName)+' ; '+CRLF

	IF EMPTY(ALLTRIM(m.ccConnectString))  && If connectstring not specified
		cCreateString = m.cCreateString + ;
			GDX_CMDMARGIN + GDX_INDENT + 'DATASOURCE "' + ALLTRIM(m.ccDataSource)    + '" ; ' + CRLF + ;
		   GDX_CMDMARGIN + GDX_INDENT + 'USERID "'     + ALLTRIM(m.ccUserId)        + '" ; ' + CRLF + ;
			GDX_CMDMARGIN + GDX_INDENT + 'PASSWORD "'   + ALLTRIM(m.ccPassword)      + '"'    + CRLF
	ELSE
		cCreateString = m.cCreateString + ;
			GDX_CMDMARGIN + GDX_INDENT + 'CONNSTRING "' + ALLTRIM(m.ccConnectString) + '"'
	ENDIF

	=WriteFile(m.hOutFile, m.cCreateString)

	*! GENERATE code to Set Connection Level Properties
	cConnectionDBSetPrefix = GDX_CMDMARGIN + [DBSetProp(']+m.cConnectionName+[', 'Connection', ]

	cConnectionProps = GDX_CMDMARGIN + '****' + CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['Asynchronous', ] + m.clAsynchronous + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['BatchMode', ] + m.clBatchMode + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['Comment', '] + m.ccComment + [')]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['DispLogin', ] + m.cnDispLogin + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['ConnectTimeOut', ] + m.cnConnectTimeOut + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['DispWarnings', ] + m.clDispWarnings + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['IdleTimeOut', ] + m.cnIdleTimeOut + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['QueryTimeOut', ] + m.cnQueryTimeOut + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
						['Transactions', ] + m.cnTransactions + [)]+ CRLF
	cConnectionProps = m.cConnectionProps + m.cConnectionDBSetPrefix + ;
					    ['Database', '] + m.ccDatabase + [')] + CRLF
					    
	=WriteFile(m.hOutFile, m.cConnectionProps)
	=WriteFile(m.hOutFile, GDX_PRCMARGIN + "ENDPROC")

RETURN

**************************************************************************
**
** Function Name: FatalAlert()
** Creation Date: 1994.12.02
** Purpose:
**
**              Place a message box to alert user of a fatal error.
**
** Parameters:
**
**      cAlert_Message - Message to display to user
**      lCleanup       - If we should try to restore environment
**
** Modification History:
**
**      1994.12.02  KRT  Added to GenDBC
**************************************************************************
PROCEDURE FatalAlert
	LPARAMETERS cAlert_Message, lCleanup

	=MessageBox(m.cAlert_Message, 16, ERROR_TITLE_LOC)

	=GenDBC_CleanUp(m.lCleanup)

	CANCEL
RETURN

**************************************************************************
**
** Function Name: GenDBC_CleanUp()
** Creation Date: 1995.03.01
** Purpose:
**
**              Restore the environment
**
** Parameters:
**
**      lCleanup - If we should try to restore tables open
**
** Modification History:
**
**      1994.03.01  KRT         Added to GenDBC
**      1997.10.01  SEA         Added code to check for existence of
**                              file before trying to reopen each table.
**************************************************************************
PROCEDURE GenDBC_CleanUp
	LPARAMETERS lCleanup

	*! Restore everything
	IF !EMPTY(m.g_cOnError)
		ON ERROR &g_cOnError
	ELSE
		ON ERROR
	ENDIF
	
	IF !EMPTY(m.g_cSetTalk)
		SET TALK &g_cSetTalk
	ENDIF
	
	IF !EMPTY(m.g_cSetDeleted)
		SET DELETED &g_cSetDeleted
	ENDIF
	
	SET EXACT &g_cSetExact
	SET SAFETY &g_cSetSafety

	IF m.g_cSetStatusBar = "OFF"
		SET STATUS BAR OFF
	ENDIF
	
	IF !EMPTY(m.g_cStatusText)
		SET MESSAGE TO (m.g_cStatusText)
	ELSE
		SET MESSAGE TO
	ENDIF
	
	SET FULLPATH &g_cFullPath
	CLOSE ALL

	IF m.lCleanUp
		IF !EMPTY(m.g_cFullDatabase) AND m.lCleanUp
			OPEN DATABASE (m.g_cFullDatabase) EXCLUSIVE
			IF m.g_nTotal_Tables_Used > 0
				FOR m.nLoop = 1 TO m.g_nTotal_Tables_Used
					*-- Make sure the file still exists!  If we had an "auto-erase"
					*-- cursor open, its file has been deleted, and can't be restored.
					*-- Thanks to Michael Colbert for finding this bug!
					IF FILE(m.g_aTables_Used(m.nLoop))
						USE (m.g_aTables_Used(m.nLoop)) AGAIN IN (m.g_aAlias_Used(m.nLoop, 2)) EXCLUSIVE ;
							ALIAS (m.g_aAlias_Used(m.nLoop, 1))
					ENDIF
				ENDFOR
			ENDIF
		ENDIF
	ENDIF
RETURN

**************************************************************************
**
** Function Name: WriteFile(, )
** Creation Date: 1994.12.02
** Purpose        :
**
**              Centralized file output routine to check for proper output
**
** Parameters:
**
**      hFileHandle - Handle of output file
**      cText       - Contents to write to file
**
** Modification History:
**
**      1994.12.02  KRT         Added to GenDBC
**************************************************************************
PROCEDURE WriteFile
	LPARAMETERS hFileHandle, cText

	nBytesSent = FPUTS(m.hFileHandle, m.cText)
	IF m.nBytesSent < LEN(m.cText)
		=FatalAlert(NO_OUTPUT_WRITTEN_LOC, .T.)
	ENDIF
RETURN

**************************************************************************
**
** Function Name: GenDBC_Error(, )
** Creation Date: 1994.12.02
** Purpose        :
**
**              Generalized Error Routine
**
** Parameters:
**
**      cMess   - Message to give user
**      nLineNo - Line Number Error Occurred
**
** Modification History:
**
**      1994.12.02  KRT         Added to GenDBC
**************************************************************************
PROCEDURE GenDBC_Error
	LPARAMETERS cMess, nLineNo

	=FatalAlert(UNRECOVERABLE_LOC + CRLF + m.cMess + CRLF + ;
				  AT_LINE_LOC + ALLTRIM(STR(m.nLineNo)), .T.)
RETURN

**************************************************************************
**
** Function Name: Stat_Message()
** Creation Date: 1994.01.08
** Purpose        :
**
**              Generalized Status Bar Progression
**
** Parameters:
**
**              None
**
** Modification History:
**
**      1994.01.08  KRT         Added to GenDBC
**************************************************************************
PROCEDURE Stat_Message
	PRIVATE ALL EXCEPT g_*
	
	nStat = m.g_nCurrentStat * (160 / g_nMax)
	SET MESSAGE TO REPLICATE("|", m.nStat) + " " + ;
		ALLTRIM(STR(INT(100 * (m.g_nCurrentStat / m.g_nMax)))) + "%"
	g_nCurrentStat = m.g_nCurrentStat + 1
RETURN


**************************************************************************
**
** Function Name: DataValue()
** Creation Date: 1997.12.07
** Purpose      : Return a representative value for parameter cDataTypeCode
**
** Parameters:
**
**     cDataTypeCode  A string holding indicating the desired
**                    data type (e.g. "C", "N", "L")
**
** Modification History:
**
**      1997.12.07  SEA         Added to GenDBCX
**************************************************************************
FUNCTION DataValue
	LPARAMETERS cDataTypeCode
	LOCAL cDataValue
	DO CASE
		CASE cDataTypeCode = "C"
			cDataValue = ["A"]
		CASE cDataTypeCode = "N" or cDataTypeCode = "Y"
			cDataValue = [0]
		CASE cDataTypeCode = "L"
			cDataValue = [.F.]
		CASE cDataTypeCode = "D"
			cDataValue = [{}]
		CASE cDataTypeCode = "T"
			cDataValue = [{}]
		OTHERWISE
			cDataValue = [null]
	ENDCASE
	RETURN cDataValue
ENDFUNC

* jme 09/02/2009 on cre la classe d'accs aux donnes pour chaque table
* S'il y a une vue _RW sur la table, on la met  jour via cette vue donc
* avec un tableupdate.
* S'il n'y a pas de vue _RW, on met  jour la table avec INSERT INTO, 
* GATHER NAME, ...
FUNCTION creation_classe_acces
  * on cre un fichier temporaire qui contiendra la ou les classes. Ce 
  * fichier sera ajout au dbut du .prg rsultat final
  * et parce que je suis moderne on utilise SET TEXTMERGE pour crer le fichier
  LOCAL lcnomfichacc, lnindice, lctable
  
  *! on ouvre de nouveau la base de donnes car on va en avoir besoin
  OPEN DATABASE (m.g_cFullDatabase) EXCLUSIVE
  lcnomfichacc = "gendbcxmacc.tmp"
  TRY 
    SET TEXTMERGE OFF
    SET TEXTMERGE TO
    SET TEXTMERGE TO (m.lcnomfichacc) NOSHOW
    * crase le fichier prcdent s'il existe
    SET TEXTMERGE ON
  CATCH
     WAIT WINDOW "impossible de crer le fichier des classes d'accs" NOWAIT NOCLEAR
     lcnomfichacc = ""
  ENDTRY
  IF _TEXT > -1
    \\* classes d'accs aux tables <<DATETIME()>>
*!*      * on commence par 'imprimer' la structure des tables, les index, etc ....
*!*      FOR  m.lnindice = 1 TO m.nTotal_Tables
*!*        lctable = ALLTRIM(CHRTRAN(aAll_Tables[ m.lnindice]," ","_"))
*!*        DO informations_table WITH m.lctable, m.lnindice
*!*      NEXT && m.lnindice = 1 TO m.nTotal_Tables
    * avant on imprimait toutes les infos tables au dbut du .prg puis on faisait les
    * classes d'accs. maintenant on met les infos de chaque table juste avant sa 
    * classe d'accs
    FOR  m.lnindice = 1 TO m.nTotal_Tables
      lctable = ALLTRIM(CHRTRAN(aAll_Tables[ m.lnindice]," ","_"))
      \
*!*        \* table <<m.lctable>>
      DO informations_table WITH m.lctable, m.lnindice
      \
      DO classe_acces_table WITH m.lctable, m.lnindice
    NEXT && m.lnindice = 1 TO m.nTotal_Tables
    \
    \* fin de la gnration des classes d'accs
    \
    \
    SET TEXTMERGE OFF
    SET TEXTMERGE TO
  ENDIF && _TEXT > -1
  CLOSE DATABASES
  RETURN m.lcnomfichacc
ENDFUNC && creation_classe_acces


PROCEDURE classe_acces_table( pctable As String, pnindice AS Number)
  LOCAL lcnomtable, lcnomclasse, lcnombdd, lcnomalias, lnindice, lnlong, ;
   llsouligne, llvuerw, lcvuerw, lnposint, lnnbint, lcsql, lnlongsql, ;
   lcparametre, lnaccu, lnnbparametres, lcaccu, lctexte, lcrequete, ;
   lnnbchamps, lnnbchampclef, llaccu, lchampcle, lnaccu2, lnlngchampclef, ;
   lcligne
  LOCAL ARRAY ltparametre(1,2), lttab(1), ltchamp(1,1), ltchampclef(1,3)
  
  * set textmerge est ON
  lcnomtable = LOWER(JUSTSTEM( m.pctable))
  lcnomclasse = "acces_"+m.lcnomtable
  lcnombdd = JUSTSTEM(m.g_cFullDatabase)
  * dans mes applis, le libell de l'alias est souvent gal aux initiales de la table
  * suivi par s2015; on essaie de l'automatiser
  lcnomalias = LEFT(m.lcnomtable,1)
  lnlong = LEN( m.lcnomtable)
  llsouligne = .F.
  FOR m.lnindice = 2 TO m.lnlong
    IF m.llsouligne
      lcnomalias = m.lcnomalias + SUBSTR(m.lcnomtable, m.lnindice, 1)
      llsouligne = .F.
      LOOP
    ELSE && m.llsouligne
      IF SUBSTR(m.lcnomtable, m.lnindice, 1) == "_"
        llsouligne = .T.
      ENDIF && SUBSTR(m.lcnomtable, m.lnindice, 1) == "_"
    ENDIF && m.llsouligne
  NEXT && m.lnindice = 2 TO m.lnlong
  IF LEN( m.lcnomalias) <2
    * si le nom de l'alias est trop court on lui ajoute le dernier caractre de lcnomtable
    lcnomalias = m.lcnomalias + RIGHT(m.lcnomtable,1)
  ENDIF && LEN( m.lcnomalias) <2
  * si on a une vue _rw, on l'utilise pour mettre  jour la table
  llvuerw = .F.
  STORE "" TO lcvuerw, ltparametre
  STORE 0 TO lnnbparametres, lnnbchamps, lnnbchampclef, lnlngchampclef
  IF m.nTotal_Views > 0
    FOR  m.lnindice = 1 TO m.nTotal_Views
      IF RIGHT(ALLTRIM(UPPER(aAll_Views[m.lnindice])), 3) == "_RW"
        * on a une vue _RW, mais concerne-t-elle la table courante ?
        lcaccu = LOWER(ALLTRIM(DBGetProp(aAll_Views[m.lnindice], 'View', 'Table')))
        * VFP met le nom de la BDD devant chaque nom de table ....
        lnaccu = AT( "!", m.lcaccu)
        IF m.lnaccu > 0
          lcaccu = SUBSTR(m.lcaccu, m.lnaccu+1)
        ENDIF && m.lnaccu > 0
*!*          \* <<aAll_Views[m.lnindice]>> table: <<m.lcaccu>> = <<m.lcnomtable>>
        lnaccu = AT(",", m.lcaccu)
        IF m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
          * une seule table et c'est la bonne
          llvuerw = .T.
          lcvuerw = ALLTRIM(LOWER(aAll_Views[m.lnindice]))
          lcrequete = g_ltsqlvue( m.lnindice)
          * il faut chercher le ou les champs clefs parce qu'on va les utiliser 
          * plus loin dans la recherche d'une nouvelle clef dans ajout()
          * on ouvre la table, puis la vue _RW pour en rcuprer la liste des champs et
          * le ou les champs qui interviennent dans la clef pour pouvoir gnrer
          * le programme de calcul de la nouvelle valeur de cette clef  partir de la
          * table clef.dbf
          * ATTENTION on est dans le programme de gnration !
          SELECT 0
          USE (m.lcnomtable) SHARED AGAIN NOUPDATE
          SELECT 0
          USE (m.lcvuerw) NODATA
          IF !EMPTY(ALIAS())
            lnnbchamps = AFIELDS(ltchamp)
          ENDIF && !EMPTY(ALIAS())
          USE IN (SELECT(m.lcvuerw))
          USE IN (SELECT(m.lcnomtable))
          EXIT
        ENDIF && m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
      ENDIF && RIGHT(ALLTRIM(UPPER(aAll_Views[m.lnindice])), 3) == "_RW"
    NEXT && m.lnindice = 1 TO m.nTotal_Views
    * si on a des vues mais aucune qui soit updatable, on peut rcuprer la liste des champs
    * sur la table elle-mme
  ENDIF && m.nTotal_Views > 0
  IF m.lnnbchamps=0
    * on n'a pas de vue ou au moins pas de vue updatable : on n'a pas pu rcuprer
    * la liste des champs. On le fait ici mais  partir de la table elle-mme
    * ATTENTION on est dans le programme de gnration !
    SELECT 0
    USE (m.lcnomtable) SHARED AGAIN NOUPDATE
    IF !EMPTY(ALIAS())
      lnnbchamps = AFIELDS(ltchamp)
    ENDIF && !EMPTY(ALIAS())
    USE IN (SELECT(m.lcnomtable))
  ENDIF && m.lnnbchamps=0
  _PRETEXT = ""
  \DEFINE CLASS <<m.lcnomclasse>> AS base_maurice && jme <<DATE()>> 
  \  * classe d'accs  la table <<LOWER(m.lcnombdd)>>!<<LOWER(m.pctable)>>
  IF m.llvuerw
    \  * avec utilisation de la vue <<lcvuerw>>
  ELSE && m.llvuerw
    \  * avec accs direct  la table (pas de vue updatable)
  ENDIF && m.llvuerw
TEXT
  ya_modif = .F.          && .T. si l'image de l'enregistrement a t modifie
  image = NULL            && image de travail 'scatter name' de banques
  image_old = NULL        && image scatter name de l'original de l'enreg de banques
                          && en comparant .image et .image_old on peut savoir s'il y a eut des
                          && modifications, abandonner les modifications en cours, ...
  alias_table = ""        && alias local de la table <<m.lcnomtable>>
  texte_erreur = ""       && message correspondant  l'erreur
  * la proprit 'erreur' est dfinie dans base_maurice
  
ENDTEXT
  IF m.llvuerw
    \  alias_vuerw = ""        && alias local de la table <<m.lcvuerw>>
  ELSE
    \  numero_ligne = 0        && RECNO() de la ligne en cours de traitement
  ENDIF && m.llvuerw
TEXT
  s2015 = ""              && rsultat de SYS(2015) : permet d'avoir la mme chaine pour
                          && toutes les vues ouvertes dans cette instance

  * FUNCTION enregistrer() .ya_modif doit tre .T.
  *                        l'image doit tre  jour.
  *                        C'est le seul endroit o on met  jour la table <<m.lcnomtable>>
  * FUNCTION ajouter()     ajouter. traite la clef. ya_modif quelconque. 
  *                        l'image doit tre  jour sauf pour la clef. 
  *                        Fonction  utiliser pour ajouter
  * FUNCTION modifier()    modifier. ya_modif quelconque.
  *                        Fonction  utiliser pour modifier
  * FUNCTION supprimer()   supprime la ligne prsente dans image. ya_modif quelconque.
  * FUNCTION chercher() && cherche une ligne et retourne l'image correspondante
  
  HIDDEN PROCEDURE Init()
    LOCAL lcnom, lccode

    DODEFAULT() && met  jour S2015
    WITH This
      .erreur = -1      && -1 non initialis, 0 non trouv, 1 ok
      .texte_erreur = "non initialis"
      .versionbase = "1.0"
      .datebase = {^<<YEAR(DATE())>>.<<MONTH(DATE())>>.<<DAY(DATE())>>}      && date de cration de cette classe
      .alias_table = "<<m.lcnomalias>>"+ .s2015
      * l'utilisation d'une chaine suivie de SYS(2015) donne un nom valide unique; de plus
      * ce nom est facilement reconnaissable dans le code de la classe
      * on peut instancier cette classe 2 fois de suite sans interference.
      .ouvertures()
      .erreur = 1 && ok l'initialisation est termine
      .texte_erreur = "ok"
    ENDWITH  && This
  ENDPROC && init
  
  HIDDEN PROCEDURE destroy 
    *  la destruction de l'objet cr  partir de cette classe on ferme les tables ouvertes
    * dans l'init. C'est un bon moyen de simuler l'environnement de donnes priv
    WITH This
      * si la dernire image traite n'est pas enregistre, on le fait (si on ne veut pas
      * il suffit de remettre ya_modif  .F. avant de dtruire l'objet
      IF .ya_modif
        .enregistrer()
      ENDIF && This.ya_modif
      * on vrifie que toutes les tables ouvertes par cet objet soient bien ferme  sa destruction
      * astuce (dans une autre classe j'ai eut  maintenir une vue ouverte aprs la destruction
      * de l'objet) : si on remet  blanc le nom d'une table ou d'une vue, on ne la referme pas
      *  la destruction de l'objet.
      .fermetures()
    ENDWITH  && This
  ENDPROC && destroy

  HIDDEN PROCEDURE ouvertures
ENDTEXT
  IF m.llvuerw
    * il faut ouvrir la vue, on recherche le ou les paramtres pour crer l'instruction LOCAL
    * avant je cherchais les paramtres en reprant les ? dans la requte SQL
    * mais cela ne me donnait pas leur type.
    * Dornavant, j'utilise parameterlist : pour chaque parametre, on un couple
    *  nom du paramtre, type mais il faut penser  le mettre  jour dans la cration de la vue
    lcparametre = ALLTRIM(DBGetProp(m.lcvuerw, 'View', 'ParameterList'))
    IF !EMPTY( m.lcparametre)
      lnaccu = ALINES( lttab, m.lcparametre, ",")
      IF m.lnaccu > 1
        lnnbparametres = INT(m.lnaccu / 2)
        DIMENSION ltparametre( m.lnnbparametres, 2)
        FOR m.lnindice = 1 TO m.lnnbparametres
          ltparametre( m.lnindice, 1) = lttab( 1 +(m.lnindice -1)*2)
          ltparametre( m.lnindice, 2) = lttab( 2 +(m.lnindice -1)*2)
*!*            \* <<m.lnindice>>: <<ltparametre( m.lnindice, 1)>>  = <<ltparametre( m.lnindice, 2)>>
        NEXT && m.lnindice = 1 TO m.lnnbparametres
      ENDIF && m.lnnbparametres >0
    ENDIF && !EMPTY( m.lcparametre)
    IF m.lnnbparametres >0
      * il y a des paramtres. Il faut crer le LOCAL et les initialiser
      * pour l'instant on ne gre pas les retours  la ligne
      * D'aprs gregory ce n'est pas obligatoire de crer le paramtre
      * pour une vue NODATA !!!
*!*        \* paramerlist = '<<m.lcparametre>>'
      \    LOCAL 
      FOR m.lnindice = 1 TO m.lnnbparametres
        lcaccu = SUBSTR(ltparametre( m.lnindice, 2), 2, 1)
        lctexte = ICASE(m.lcaccu=="I", "Integer", m.lcaccu=="C", "String", ;
         m.lcaccu=="N", "Number", m.lcaccu=="D", "Date", m.lcaccu=="T", "dateTime", ;
          m.lcaccu=="L", "Logical", "unknown")
        \\<<ltparametre( m.lnindice, 1)>> AS <<m.lctexte>>
        \\<<IIF(m.lnindice<m.lnnbparametres, ", ", "")>>
      NEXT && m.lnindice = 1 TO m.lnnbparametres
      \    * d'aprs Gregory, c'est inutile ...
    ENDIF && m.lnnbparametres >0
  ENDIF && m.llvuerw
TEXT
    WITH This
      SELECT 0
      USE <<LOWER(m.lcnombdd)>>!<<LOWER(m.pctable)>> AGAIN SHARED ALIAS (.alias_table) IN 0
      SCATTER NAME .image BLANK && initialisation des images
      SCATTER NAME .image_old BLANK
ENDTEXT
  IF m.llvuerw
    * on va ouvrir la vue, il faut initialiser les paramtres
    * Gregory dit que ce n'est pas obligatoire avec une vue NODATA
    IF m.lnnbparametres >0
      FOR m.lnindice = 1 TO m.lnnbparametres      
        lcaccu = SUBSTR(ltparametre( m.lnindice, 2), 2, 1)
        DO CASE 
        CASE m.lcaccu $ "IN"
          lctexte = "0"
        CASE m.lcaccu $ "C"
          lctexte = '""'
        CASE m.lcaccu $ "DT"
          lctexte = "{}"
        CASE m.lcaccu $ "L"
          lctexte = ".F."
        OTHERWISE
          lctexte = ".NULL."
        ENDCASE
        \      <<ltparametre( m.lnindice, 1)>> = <<m.lctexte>>
      NEXT && m.lnindice = 1 TO m.lnnbparametres
*!*  Return value  Data type  
*!*  C Character, Memo, Varchar, Varchar (Binary)
*!*  D Date      G General      L Logical      N Numeric, Float, Double, or Integer
*!*  O Object    Q Blob, Varbinary    T DateTime    U Unknown or variable does not exist
*!*  X Null      Y Currency
 
    ENDIF && m.lnnbparametres >0
    \      .alias_vuerw = "a"+ .alias_table
    \      SELECT 0
    \      USE <<m.lcvuerw>> ALIAS (.alias_vuerw) NODATA
    lnaccu = MEMLINES(m.lcrequete)
    FOR  m.lnindice = 1 TO m.lnaccu
      \      *!<<SPACE(m.lnindice)>><<ALLTRIM(MLINE(m.lcrequete, m.lnindice))>>
    NEXT && m.lnindice = 1 TO m.lnaccu
  ELSE
    \      .numero_ligne = 0
  ENDIF && m.llvuerw
TEXT
    ENDWITH  && This
  ENDPROC && ouvertures
  
  HIDDEN PROCEDURE fermetures
    * si par hasard, on a besoin de conserver la table et/ou les vues ouvertes aprs le
    * destroy de l'objet, il n'y a qu' remettre  blanc la proprit correspondante
    WITH This
ENDTEXT
  IF m.llvuerw
    \      IF !EMPTY(.alias_vuerw)
    \        USE IN (SELECT(.alias_vuerw))
    \      ENDIF
  ENDIF && m.llvuerw
TEXT
      IF !EMPTY(.alias_table)
        USE IN (SELECT(.alias_table))
      ENDIF
    ENDWITH  && This
  ENDPROC && fermetures

ENDTEXT
  * on gnre la fonction enregistrer de 2 manires diffrentes selon que
  * l'on a une vue updatable ou si on travaille directement sur la table
  IF m.llvuerw
    * on rcupre le ou les champs clef
    FOR m.lnindice = 1 TO m.lnnbchamps
      lcaccu = m.lcvuerw+"."+ltchamp(m.lnindice, 1)
      IF DBGETPROP( m.lcaccu, 'Field', 'KeyField')
        lnnbchampclef = m.lnnbchampclef + 1
        IF m.lnnbchampclef > 1
          DIMENSION ltchampclef(m.lnnbchampclef, 3)
        ENDIF && m.lnnbchampclef > 1
        ltchampclef(m.lnnbchampclef,1) = LOWER(ltchamp(m.lnindice, 1))
        ltchampclef(m.lnnbchampclef, 2) = ;
         DBGETPROP( m.lcaccu, 'Field', 'Updatable')
        ltchampclef(m.lnnbchampclef, 3) = m.lnindice
      ENDIF && DBGETPROP( m.lcvuerw+"!"+ltchamp(m.lnindice, 1), 'Field', 'KeyField')
    NEXT && m.lnindice = 1 TO m.lnnbchamps
    \*
    \* nombre de champ clef: <<m.lnnbchampclef>>
    FOR m.lnindice = 1 TO m.lnnbchampclef
      \*    <<ltchampclef(m.lnindice, 1)>> <<ltchampclef(m.lnindice, 2)>>
    NEXT && m.lnindice = 1 TO m.lnnbchampclef
    * on indique si, par accident, un champ clef n'est pas updatable
    llaccu = .F.
    FOR m.lnindice = 1 TO m.lnnbchampclef
      IF ! ltchampclef(m.lnindice, 2)
        IF m.llaccu = .F.
          \* ATTENTION champ clef non Updatable !! :
          llaccu = .T.
        ENDIF && m.llaccu = .F.
        \*    <<ltchampclef(m.lnindice, 1)>>
      ENDIF && ! ltchampclef(m.lnindice, 2)
    NEXT && m.lnindice = 1 TO m.lnnbchampclef
    * s'il y a plusieurs champs qui interviennent dans la clef, ils sont forcment
    * de type Character (les seuls qui soient concatnables)
    IF m.lnnbchampclef>1
      llaccu = .F.
      lnaccu = 0
      FOR m.lnindice = 1 TO m.lnnbchampclef
        lnaccu2 = ltchampclef(m.lnindice, 3)
        IF ! ltchamp( m.lnaccu2, 2) == "C"
          IF m.llaccu = .F.
            \* ATTENTION plusieurs champs clef non Character !! :
            llaccu = .T.
          ENDIF && m.llaccu = .F.
          \*    <<ltchampclef(m.lnindice, 1)>> <<ltchamp(m.lnaccu2, 2)>>
        ELSE
          * la somme des longueurs des champs entrant dans la clef doit tre 2 ou 4
          * (peut pas tre 1 ! il y a plusieurs champs !!)
          lnaccu = m.lnaccu + ltchamp( m.lnaccu2, 3)
        ENDIF && ! ltchampclef(m.lnindice, 2)
      NEXT && m.lnindice = 1 TO m.lnnbchampclef
      IF ! INLIST(m.lnaccu2, 2, 4)
        \* ATTENTION les champs clefs doivent tre de longueur 2 ou 4 au total
        lnlngchampclef = 4 && scurit
      ELSE
        lnlngchampclef = m.lnaccu2
      ENDIF && ! INLIST(m.lnaccu2, 2, 4)
    ELSE
      * si le champ clef est de type character, sa longueur ne peut tre que 1, 2 ou 4
      lnaccu = ltchampclef(1, 3)
      IF ltchamp( m.lnaccu, 2) == "C"
        IF ! INLIST(ltchamp( m.lnaccu, 3), 1, 2, 4)
          \* ATTENTION le champ <<LOWER(ltchamp( m.lnaccu, 1))>> doit tre de longueur 1, 2 ou 4
          lnlngchampclef = 4 && scurit
        ELSE
          lnlngchampclef = ltchamp( m.lnaccu, 3)
        ENDIF
      ENDIF && ltchamp(m.lnaccu, 2) == "C"
    ENDIF && m.lnnbchampclef>1
    
*!*  One row is created for each field in the table.
*!*  Column number  Field info  Data type  
*!*  1  Field name  Character
*!*  2  Field type: 
*!*     C = Character     Y = Currency
*!*     D = Date          T = DateTime
*!*     B = Double        F = Float
*!*     G = General       I = Integer
*!*     L = Logical       M = Memo
*!*     N = Numeric       Q = Varbinary
*!*     V = Varchar and Varchar (Binary)
*!*     W = Blob  Character
*!*  3 Field width  Numeric
*!*  4 Decimal places  Numeric
*!*  5 Null values allowed  Logical
*!*  6 Code page translation not allowed  Logical
*!*  7 Field validation expression Character
*!*  8 Field validation text Character
*!*  9 Field default value Character

TEXT
  FUNCTION enregistrer( pnoperation AS Number)
    * pnoperation : 1 ajout, 2 modification, 3 suppression
    * l'image doit tre  jour y compris la clef
    * l'enregistrement se fait en deux temps :
    *  1) on met  jour la vue <<m.lcvuerw>>  partir de l'image
    *  2) on fait un tableupdate de cette vue
    LOCAL llresult, lnaccu, lccode
    
    IF This.ya_modif
      llresult = .F.
      IF m.pnoperation = 1
        * on ajoute; la clef a dj t mise  jour
        IF This.erreur >0 && scurit
          * la vue <<m.lcvuerw>> a t ouverte dans This.ouvertures; elle peut avoir
          * dj un enregistrement qu'il faut alors blanchir
          SELECT (m.lcvuerw)
          IF RECCOUNT() = 0
            APPEND BLANK
          ELSE && RECCOUNT() = 0
            IF RECCOUNT() = 1
              GO TOP
              BLANK
            ELSE && RECCOUNT() = 1
              * normalement on ne devrait jamais venir ici
              This.erreur = -201
              This.texte_erreur = "la vue <<m.lcvuerw>> ne devrait pas avoir plusieurs lignes"
              ZAP  && a-t-on le droit de ZAPper une vue ?
              APPEND BLANK
            ENDIF && RECCOUNT() = 1
          ENDIF && RECCOUNT() = 0
        ENDIF && This.erreur >0
      ELSE
        IF m.pnoperation = 2
          * modification
          SELECT (m.lcvuerw)
          IF RECCOUNT() <> 1
            This.erreur = -200
            This.texte_erreur = "la vue <<m.lcvuerw>> devrait avoir une et une seule ligne en modification"
            ZAP  && a-t-on le droit de ZAPper une vue ?
            APPEND BLANK
          ENDIF && RECCOUNT() = 1
        ELSE && m.pnoperation = 2
          IF m.pnoperation = 3
            * suppression
            SELECT (m.lcvuerw)
            IF RECCOUNT() <> 1
              This.erreur = -202
              This.texte_erreur = "la vue <<m.lcvuerw>> devrait avoir une et une seule ligne en suppression"
              ZAP  && a-t-on le droit de ZAPper une vue ?
              APPEND BLANK
            ENDIF && RECCOUNT() = 1
          ELSE
            This.erreur = -203 && opration inconnue
            This.texte_erreur = "opration inconnue:"+ STR(m.pnoperation,3)
          ENDIF && m.pnoperation = 3
        ENDIF && m.pnoperation = 2
      ENDIF && <<m.lcaccu>>
      IF This.erreur >0
        LOCAL ARRAY lterr2(1)
        GATHER NAME This.image MEMO
ENDTEXT
    * dans mes tables, j'ai souvent un champ DATCREAT D ou CREAT T qui
    * contient la date de dernire mise  jour. Il est souvent ' la fin'
    * s'il existe, on le traite
    llaccu = .F.
    m.lnindice = m.lnnbchamps
    DO WHILE m.lnindice > 1 AND m.llaccu = .F.
      IF "CREAT" $ UPPER(ltchamp(m.lnindice, 1)) AND ;
       ltchamp(m.lnindice, 2) $ "DT"
        * on recherche,  partir de la fin, le premier champ dont le nom contient
        * 'creat' et qui est du type date ou datetime
        llaccu = .T.
      ELSE
        m.lnindice = m.lnindice - 1
      ENDIF      
    ENDDO && m.lnindice > 1
    IF m.llaccu
      \        REPLACE <<LOWER(ltchamp(m.lnindice, 1))>> WITH 
      \\<<IIF(ltchamp(m.lnindice, 2) == "D", "DATE()", "DATETIME()")>>
    ENDIF && m.llacu
TEXT
        * on enregistre la vue dans la table
        * 0 : un seul enregistrement  .F. : on ne force pas en cas de conflit
        * lterr2 tableau pour les recno()  problmes
        IF TABLEUPDATE(0, .F., this.alias_vuerw, lterr2)
          This.ya_modif = .F. && jme 03/02/2006
          llresult = .T.
        ELSE
          * on a un problme avec le tableupdate. L'oprateur va en tre averti par la fonction
          * gnrale traitement_erreur_tableupdate. Mais comme il ne pensera pas  noter
          * le message pour me le dire, je moucharde l'vnement pour tre capable, plus
          * tard de retrouver le problme.
          This.erreur = -10 && utilisable dans la 'procdure appelante'
          This.texte_erreur = "chec TABLEUPDATE <<m.lcnomtable>>"
          LOCAL lcaccu
          LOCAL ARRAY ltparm(10)
          ltparm(1) = "<<m.lcnomtable>>"
          STORE "" TO ltparm(2), ltparm(3), ltparm(4), ltparm(5), ltparm(6), ltparm(7)
          lcaccu = traitement_erreur_tableupdate( lterr2, ltparm) && dans standard_maurice.prg
          * le tablerevert () est fait dans traitement_erreur_tableupdate
        ENDIF && llok
      ENDIF && m.llresult
      RETURN m.llresult
    ELSE
      RETURN .T.
    ENDIF && This.ya_modif
  ENDPROC
  
  PROCEDURE nouvelle_clef && 11/02/2009
    * va chercher une nouvelle clef pour la table <<m.lcnomtable>> dans la table clefs
    * si on vient ici, dans le programme gnr, c'est qu'il n'y a qu'un seul champ clef
    LOCAL lnaccu
    
ENDTEXT
    lchampcle = LOWER(ltchamp(ltchampclef(1, 3), 1))
    IF ltchamp(ltchampclef(1, 3), 2) == "C"
      \    this.image.<<m.lchampcle>> = BINTOC(0,<<m.lnlngchampclef>>)
      lcaccu = "BINTOC( m.lnaccu,"+STR(m.lnlngchampclef,2)+ ")"
    ELSE
      \    this.image.<<m.lchampcle>> = 0
      lcaccu = "m.lnaccu"
    ENDIF && ltchamp(ltchampclef(1, 3), 2) == "C"
TEXT
    * pour l'instant on travaille directement sur 'clefs'
    * plus tard, il faudra crer une vue bufferise ligne pessimiste
    IF USED("clefs")
      SELECT clefs
      IF SEEK("<<m.lcnomtable>>", "clefs", "alias")
        IF RLOCK("clefs")
          lnaccu = clefs.derclefn + 1
          REPLACE clefs.derclefn WITH m.lnaccu, clefs.date WITH DATETIME()
          UNLOCK IN clefs
          STORE <<m.lcaccu>> TO this.image.<<m.lchampcle>>
          This.erreur = 1 && ok
          This.texte_erreur = "ok"
        ELSE
          This.erreur = -100
          This.texte_erreur = "impossible de rserver CLEFS"
        ENDIF && RLOCK()
      ELSE
        This.erreur = -101
        This.texte_erreur = "pas de ligne pour <<m.lcnomtable>> dans clefs"
      ENDIF && SEEK("<<m.lcnomtable>>")
    ELSE
      This.erreur = -102
      This.texte_erreur = "table des clefs non ouverte"
    ENDIF && USED(clefs)
  ENDFUNC && nouvelle_clef

ENDTEXT
    
  ELSE && m.llvuerw

TEXT
  FUNCTION enregistrer( pnoperation AS Number)
    * en cours de programmation dans gendbcxm
    * ici on enregistre directement This.image dans la table
    * pnoperation : 1 ajout, 2 modification, 3 suppression
    * l'image doit tre  jour y compris la clef
    LOCAL llresult, lnaccu, lccode
    
    IF This.ya_modif
      llresult = .F.
      IF This.erreur >0 && scurit
        SELECT This.alias_table
        IF m.pnoperation <3
          IF m.pnoperation = 1
            * on ajoute; la clef a dj t mise  jour
            APPEND BLANK
          ELSE && m.pnoperation = 1
            * modification
          ENDIF && m.pnoperation = 1
          * mettre ici les ventuelles mises  jour de champs
ENDTEXT          
    * dans mes tables, j'ai souvent un champ DATCREAT D ou CREAT T qui
    * contient la date de dernire mise  jour. Il est souvent ' la fin'
    * s'il existe, on le traite
    llaccu = .F.
    m.lnindice = m.lnnbchamps
    DO WHILE m.lnindice > 1 AND m.llaccu = .F.
      IF "CREAT" $ UPPER(ltchamp(m.lnindice, 1)) AND ;
       ltchamp(m.lnindice, 2) $ "DT"
        * on recherche,  partir de la fin, le premier champ dont le nom contient
        * 'creat' et qui est du type date ou datetime
        llaccu = .T.
      ELSE
        m.lnindice = m.lnindice - 1
      ENDIF      
    ENDDO && m.lnindice > 1
    IF m.llaccu
      \          This.image.<<LOWER(ltchamp(m.lnindice, 1))>> = 
      \\<<IIF(ltchamp(m.lnindice, 2) == "D", "DATE()", "DATETIME()")>>
    ENDIF && m.llacu
TEXT
          TRY 
            GATHER NAME This.image MEMO
            llresult = .T.
          CATCH
            This.erreur = -250
            This.texte_erreur = "chec GATHER NAME sur <<m.lcnomtable>>"
          ENDTRY
        ELSE && m.pnoperation  <3
          IF m.pnoperation = 3
            * suppression
            
          ELSE
            This.erreur = -203 && opration inconnue
            This.texte_erreur = "opration inconnue:" + STR(m.pnoperation,3)
          ENDIF && m.pnoperation = 3
        ENDIF && m.pnoperation  <3
      ELSE && This.erreur>0 
        
      ENDIF && This.erreur>0 
      RETURN m.llresult
    ELSE
      RETURN .T.
    ENDIF && This.ya_modif
  ENDPROC && enregistrer

  PROCEDURE nouvelle_clef && 13/02/2009
    * va chercher une nouvelle clef pour la table <<m.lcnomtable>> dans la table clefs
    * Il faut vrifier, dans le programme gnr que, si on vient ici, c'est qu'il 
    * n'y a qu'un seul champ clef
    LOCAL lnaccu
    * comme on n'a pas de vue, on est incapable de savoir quel est le champ clef !!
    * dans le programme gnr, il faudra remplacer ?? par le nom du champ
    * pour l'instant on travaille directement sur 'clefs'
    IF USED("clefs")
      SELECT clefs
      IF SEEK("<<m.lcnomtable>>", "clefs", "alias")
        IF RLOCK("clefs")
          lnaccu = clefs.derclefn + 1
          REPLACE clefs.derclefn WITH m.lnaccu, clefs.date WITH DATETIME()
          UNLOCK IN clefs
          * il faut peut-tre transformer la clef numrique en caractres ...
          * penser  remplacer ?? par le champ clef
          STORE m.lnaccu TO this.image.??
          This.erreur = 1 && ok
          This.texte_erreur = "ok"
        ELSE
          This.erreur = -100
          This.texte_erreur = "impossible de rserver CLEFS pour <<m.lcnomtable>>"
        ENDIF && RLOCK()
      ELSE
        This.erreur = -101
        This.texte_erreur = "pas de ligne pour <<m.lcnomtable>> dans clefs"
      ENDIF && SEEK("<<m.lcnomtable>>")
    ELSE
      This.erreur = -102
      This.texte_erreur = "table des clefs non ouverte"
    ENDIF && USED(clefs)
  ENDFUNC && nouvelle_clef

ENDTEXT
    
  ENDIF && m.llvuerw
  *
  * traitement de la suppression
  * si la clef est un Entier, on 'supprime' la ligne en passant la clef en ngatif
  * (et on met un commentaire si aucune vue ne filtre la clef positive)
  * sinon
  *   
  IF m.llvuerw
TEXT
  FUNCTION supprimer
    * l'image doit tre  jour (mais c'est ici que l'on passe la clef en ngatif)
    * l'enregistrement se fait en trois temps :
    *  1) on met  jour la vue <<m.lcvuerw>>  partir de l'image
    *  2) on passe la clef en ngatif (plus tard on essaiera de mettre en place un
    *     moyen de rcuprer une ligne supprime lorsque l'on crer une nouvelle
    *     ligne (recyclage des lignes supprimes)
    *  3) on fait un tableupdate de cette vue
  
ENDTEXT

    IF ltchamp(ltchampclef(1, 3), 2) == "C"
      \    LOCAL llresult, lccode, lcnoucode
      \
      \    WITH This
      \      * provisoirement, on dtaille l'opration pour faciliter le deboguage ventuel
      \      lccode = .image.<<m.lchampcle>>
      \      * ABS : scurit des fois que la ligne soit dj efface
      \      lcnoucode = BINTOC( - ABS(CTOBIN( m.lccode)),<<m.lnlngchampclef>>)
      \      .image.<<m.lchampcle>> = m.lcnoucode
    ELSE
      \    LOCAL llresult
      \
      \    WITH This
      \      * ABS : scurit des fois que la ligne soit dj efface
      \      .image.<<m.lchampcle>> = -ABS(.image.<<m.lchampcle>>)
    ENDIF && ltchamp(ltchampclef(1, 3), 2) == "C"
        
TEXT
      .ya_modif = .T.
      llresult = .enregistrer(3)
      IF m.llresult
        * mettre ici le code ventuel en cas de russite du tableupdate
      ELSE
        * mettre ici le code ventuel en cas d'chec du tableupdate
      ENDIF && m.llresult
    ENDWITH && This
    RETURN m.llresult
  ENDFUNC && supprimer
ENDTEXT

  ELSE && m.llvuerw

TEXT
  FUNCTION supprimer
    * non encore programm (effacement direct sans vue)
  ENDFUNC && supprimer  
ENDTEXT

  ENDIF && m.llvuerw
      
TEXT
  
  FUNCTION ajouter()
    * on ajoute dans la table <<m.lcnomtable>>
    WITH This
      * mettre ici les mises  jour ventuelles des champs calculs
      * dans ma BDD, j'ai une table "clefs" qui contient une ligne par table; chaque ligne
      * contient la valeur de la dernire clef utilise pour la creation d'une ligne
      * dans la table correspondante. Cette table clef sera aussi utilise, plus tard, 
      * pour tenter de rcuprer les enregistrements marqus pour l'effacement 
      * (normalement en mettant un code ngatif).
ENDTEXT
    IF m.llvuerw
      IF m.lnnbchampclef = 1
        \      This.nouvelle_clef()
      ELSE && m.lnnbchampclef = 1
        \      *
        \      * la vue <<m.lcvuerw>> a plusieurs champs clef. Il faut calculer la nouvelle
        \      * valeur de ces champs ici
      ENDIF && m.lnnbchampclef = 1
    ELSE && m.llvuerw
      \      *
      \      * Il faut calculer la nouvelle valeur du ou des champs clef ici
    ENDIF && m.llvuerw
TEXT
      IF This.erreur > 0
        .ya_modif = .T.
        IF .enregistrer(1) && 1 = ajout
          * mettre ici le code ventuel pour les traitements aprs enregistrement
          RETURN .T.
        ELSE
          RETURN .F.
        ENDIF && .enregistrer()
      ELSE && This.erreur > 0
        RETURN .F. && impossible d'obtenir une nouvelle clef
      ENDIF && This.erreur > 0
    ENDWITH  && This
  ENDFUNC &&* ajouter
  
  FUNCTION modifier()
    * on modifie dans la table <<m.lcnomtable>>
    WITH This
      * mettre ici les mises  jour ventuelles des champs calculs
      .ya_modif = .T.
      IF .enregistrer(2)
        * mettre ici le code ventuel pour les traitements aprs enregistrement
        RETURN .T.
      ELSE
        RETURN .F.
      ENDIF && .enregistrer()
    ENDWITH  && This
  ENDFUNC &&* modifier
  
ENDTEXT
  
  IF m.lnnbchampclef > 0
    * on a au moins un champ clef, on peut faire une fonction 'chercher'
    * si on a une vue, on l'utilise sinon on fait un seek
    * si on a une vue, il faut dclarer puis mettre  jour le paramtre
    IF m.lnnbparametres > 0
      lcligne = "LOCAL "
      FOR m.lnindice = 1 TO m.lnnbparametres
        lcaccu = SUBSTR(ltparametre( m.lnindice, 2), 2, 1)
        lctexte = ICASE(m.lcaccu=="I", "Integer", m.lcaccu=="C", "String", ;
         m.lcaccu=="N", "Number", m.lcaccu=="D", "Date", m.lcaccu=="T", "dateTime", ;
          m.lcaccu=="L", "Logical", "unknown")
        lcligne = m.lcligne + ltparametre( m.lnindice, 1) + " AS "+ m.lctexte+ ;
         IIF(m.lnindice < m.lnnbparametres, ", ", "")
      NEXT && m.lnindice = 1 TO m.lnnbparametres
    ELSE && m.lnnbparametres > 0
      lcligne = "* ATTENTION : pas de paramtre dans la vue "+ m.lcvuerw
      * glut : en scurit, on met un message d'erreur dans ltparametre !!
      DIMENSION ltparametre(1,2)
      ltparametre(1,1) = "???!!??"
      ltparametre(1,2) = "?"
    ENDIF && m.lnnbparametres > 0
    DO CASE 
    CASE m.lnnbchampclef = 1
      * si on n'a qu'un seul paramtre, il peut tre de diffrents types
      * et comme j'aime bien que le nom du paramtre de la fonction indique
      * son type (avec la deuxime lettre) cela complique un peu les choses
      * mais bon !
      lcaccu = ltchamp(ltchampclef(1, 3), 2)
      DO CASE 
      CASE m.lcaccu == "C"
        lctexte = "pcclef"
        \  FUNCTION chercher( pcclef AS String)
        \    <<m.lcligne>>
        \
        \    <<ltparametre( 1, 1)>> = m.pcclef
      CASE m.lcaccu == "N"
        lctexte = "pnclef"
        \  FUNCTION chercher( pnclef AS Number)
        \    <<m.lcligne>>
        \
        \    <<ltparametre( 1, 1)>> = m.pnclef
      CASE m.lcaccu == "I"
        lctexte = "peclef"
        \  FUNCTION chercher( peclef AS Integer)
        \    <<m.lcligne>>
        \
        \    <<ltparametre( 1, 1)>> = m.peclef
      OTHERWISE && type inconnu !!
        lctexte = "puclef"
        \  FUNCTION chercher( puclef)
        \    <<m.lcligne>>
        \
        \    <<ltparametre( 1, 1)>> = m.puclef
      ENDCASE
      * si on a plus d'un paramtre pour la vue et un seul pour la fonction,
      * il y a un problme ! on le signale
      IF m.lnnbparametres > 1
        \    * il y a plus de paramtres dans la vue que dans la fonction !!
        \    * que fait-on de 
        FOR m.lnindice = 2 TO m.lnnbparametres
          \\<<ltparametre( m.lnindice, 1)>><<IIF(m.lnindice < m.lnnbparametres, ", ", "")>>
        NEXT && m.lnindice = 1 TO m.lnnbparametres
        \\ ?
      ENDIF && m.lnnbparametres > 1
      
    CASE m.lnnbchampclef = 2
      * les deux paramtres sont forcment de type caractres (du moins 
      * pour l'instant; plus tard on pourra voir s'il faut faire des
      * CTOBIN(BINTOC()+BINTOC()) !!)
      * mais non, je ne suis pas compliqu !!!
      \  FUNCTION chercher(pcclef1 As String, pcclef2 AS String)
      \    <<m.lcligne>>
      \
      DO CASE 
      CASE m.lnnbparametres = 0
        * on ne fait rien, l'erreur est dj signale
      CASE m.lnnbparametres = 1
        * il faut concatner les paramtres de la fonction dans le paramtre de la vue
        \    <<ltparametre( 1, 1)>> = m.pcclef1 + m.pcclef2
      CASE m.lnnbparametres = 2
        * chaque paramtre de la fonction va dans le paramtre correspondant
        * dans la vue
        \    <<ltparametre( 1, 1)>> = m.pcclef1
        \    <<ltparametre( 2, 1)>> = m.pcclef2
      OTHERWISE
        \    * dmerdez-vous avec les paramtres !
        \    * 
        FOR m.lnindice = 1 TO m.lnnbparametres
          \\<<ltparametre( m.lnindice, 1)>><<IIF(m.lnindice < m.lnnbparametres, ", ", "")>>
        NEXT && m.lnindice = 1 TO m.lnnbparametres
      ENDCASE

    CASE m.lnnbchampclef = 3
      * les trois paramtres sont forcment de type caractres
      \  FUNCTION chercher(pcclef1 As String, pcclef2 AS String, pcclef3 As String)
      \    <<m.lcligne>>
      \
      DO CASE 
      CASE m.lnnbparametres = 0
        * on ne fait rien, l'erreur est dj signale
      CASE m.lnnbparametres = 1
        * il faut concatner les paramtres de la fonction dans le paramtre de la vue
        \    <<ltparametre( 1, 1)>> = m.pcclef1 + m.pcclef2 + m.pcclef3
      CASE m.lnnbparametres = 3
        * chaque paramtre de la fonction va dans le paramtre correspondant
        * dans la vue
        \    <<ltparametre( 1, 1)>> = m.pcclef1
        \    <<ltparametre( 2, 1)>> = m.pcclef2
        \    <<ltparametre( 3, 1)>> = m.pcclef3
      OTHERWISE  && valable aussi si m.lnnbparametres = 2 !
        \    * dmerdez-vous avec les paramtres !
        \    * 
        FOR m.lnindice = 1 TO m.lnnbparametres
          \\<<ltparametre( m.lnindice, 1)>><<IIF(m.lnindice < m.lnnbparametres, ", ", "")>>
        NEXT && m.lnindice = 1 TO m.lnnbparametres
      ENDCASE
    OTHERWISE
      * on ne passe qu'un seul paramtre qui est la concatnation des n paramtres !
      \  FUNCTION chercher(pcclef As String)
      \    * pcclef doit tre la concatnation des <<m.lnnbchampclef>> paramtres
      \    <<m.lcligne>>
      \
      \    * dmerdez-vous avec les paramtres !
      \    * 
      FOR m.lnindice = 1 TO m.lnnbparametres
        \\<<ltparametre( m.lnindice, 1)>><<IIF(m.lnindice < m.lnnbparametres, ", ", "")>>
      NEXT && m.lnindice = 1 TO m.lnnbparametres
    ENDCASE
TEXT
    SELECT <<m.lcvuerw>>
    REQUERY()
    IF RECCOUNT() > 0
      GATHER NAME This.image MEMO
      GATHER NAME This.image_old MEMO
      This.yamodif = .F.
      This.erreur = 1
      RETURN .T.
    ELSE && RECCOUNT() > 0
      GATHER NAME This.image MEMO BLANK
      GATHER NAME This.image_old MEMO BLANK
      This.erreur = -400
      This.texte_erreur = "aucune ligne trouve"
      This.yamodif = .F.
      RETURN .F.
    ENDIF && RECCOUNT() > 0
  ENDFUNC && chercher  
ENDTEXT
    
  ELSE && m.lnnbchampclef > 0
    \  * pas de champ clef => pas de fonction chercher !
  ENDIF && m.lnnbchampclef > 0
  \
  \ENDDEFINE && <<m.lcnomclasse>>
  \
  RETURN
ENDPROC && classe_acces_table


PROCEDURE informations_table( pctable As String, pnindice AS Number)
  * on 'imprime' la structure de la table en essayant de mettre toutes 
  * les infos disponible dans la BDD puis on liste les index
  LOCAL lcnomtable, lnindice, lnlong, lnaccu, lcaccu, lnnbchamp, lntotal, lnlong2, ;
   llaccu, lnnombre, lcnomvue, llupdatable, lnligne, lni
  LOCAL ARRAY ltstruct(1,1), lttag(1,1), lttab(1)
  
  lcnomtable = LOWER(JUSTSTEM( m.pctable))
  \* Informations sur la table <<m.lcnomtable>>
  SELECT 0
  USE (m.lcnomtable) SHARED AGAIN NOUPDATE
  lnnbchamp = AFIELDS( ltstruct)
  \*   RECCOUNT = <<RECCOUNT()>>, nbre de champs = <<m.lnnbchamp>>
  \*   Dernire mise  jour: <<LUPDATE()>>, page de code: <<CPDBF()>>
  IF !EMPTY( ltstruct( 1, 12))
    \*   Nom long: <<ltstruct( 1, 12)>>
  ENDIF && !EMPTY( ltstruct( 1, 10))
  IF !EMPTY( ltstruct( 1, 10))
    \*   Rgle validation table: <<ltstruct( 1, 10)>>
  ENDIF && !EMPTY( ltstruct( 1, 10))
  IF !EMPTY( ltstruct( 1, 11))
    \*   Texte validation table: <<ltstruct( 1, 11)>>
  ENDIF && !EMPTY( ltstruct( 1, 10))
  IF EMPTY( ltstruct( 1, 13)) AND EMPTY( ltstruct( 1, 14)) AND EMPTY( ltstruct( 1, 15))
    \*   Pas de trigger
  ELSE && EMPTY( ltstruct( 1, 13)) AND EMPTY( ltstruct( 1, 14)) AND EMPTY( ltstruct( 1, 15))
    IF !EMPTY( ltstruct( 1, 13))
      \*   Trigger Insert: <<ltstruct( 1, 13)>>
    ENDIF && !EMPTY( ltstruct( 1, 13))
    IF !EMPTY( ltstruct( 1, 14))
      \*   Trigger Update: <<ltstruct( 1, 14)>>
    ENDIF && !EMPTY( ltstruct( 1, 14))
    IF !EMPTY( ltstruct( 1, 15))
      \*   Trigger Update: <<ltstruct( 1, 15)>>
    ENDIF && !EMPTY( ltstruct( 1, 15))
  ENDIF && EMPTY( ltstruct( 1, 13)) AND EMPTY( ltstruct( 1, 14)) AND EMPTY( ltstruct( 1, 15))
  IF !EMPTY( ltstruct( 1, 16))
    \*   Note: <<ltstruct( 1, 16)>>
  ENDIF && !EMPTY( ltstruct( 1, 10))
  IF ltstruct( 1, 17) <> 0
    \*  nextvalue for autoincrementing <<ltstruct( 1, 17)>>
  ENDIF && ltstruct( 1, 17) > 0
  IF ltstruct( 1, 18) <> 0
    \*  Step for autoincrementing <<ltstruct( 1, 18)>>
  ENDIF && ltstruct( 1, 18) > 0
  * pour bien aligner les colonnes je cherche la longueur la plus grande
  lnlong = 0
  FOR  m.lnindice = 1 TO m.lnnbchamp
    lnlong = MAX(LEN(ltstruct( m.lnindice, 1)), m.lnlong)
  NEXT && m.lnindice = 1 TO m.lnnbchamp
  lnlong = m.lnlong + 1
  lntotal = 0
  FOR m.lnindice = 1 TO m.lnnbchamp
    \*  <<STR(m.lnindice,3)>> <<PADR(ltstruct( m.lnindice, 1), m.lnlong)>> 
    \\<<PADR(lettre_vers_type(ltstruct( m.lnindice, 2)), 10)>> 
    \\<<STR(ltstruct( m.lnindice, 3),3)>> 
    \\<<IIF(ltstruct( m.lnindice, 4)>0, STR(ltstruct( m.lnindice, 4),2), "  ")>> 
    \\<<IIF(ltstruct( m.lnindice, 5), " NULL   ", "NOT NULL")>> 
    \\<<IIF(!EMPTY(ltstruct( m.lnindice, 9)), "default: "+ltstruct( m.lnindice, 9), " ")>>
    \\<<IIF(ltstruct( m.lnindice, 6), "NO CPTRANS", "")>>
    lntotal = m.lntotal + ltstruct( m.lnindice, 3)
  NEXT && m.lnindice = 1 TO m.lnnbchamp
  \*<<SPACE(m.lnlong+10)>>Total: <<STR(m.lntotal,4)>>
  * on regarde s'il y a des rgles de validation et/ou des textes au niveau des champs
  llaccu = .F.
  FOR m.lnindice = 1 TO m.lnnbchamp
    llaccu = IIF(!EMPTY(ltstruct( m.lnindice, 7)) OR !EMPTY(ltstruct( m.lnindice, 8)), .T., m.llaccu)
  NEXT && m.lnindice = 1 TO m.lnnbchamp
  IF m.llaccu
    * il y a des rgles de validation au niveau des champs
    \*  rgles de validation champ, textes :
    FOR m.lnindice = 1 TO m.lnnbchamp
      IF !EMPTY(ltstruct( m.lnindice, 7)) OR !EMPTY(ltstruct( m.lnindice, 8))
        IF !EMPTY(ltstruct( m.lnindice, 7))
          \*  <<STR(m.lnindice,3)>> <<PADR(ltstruct( m.lnindice, 1), m.lnlong)>> : 
          \\<<ltstruct( m.lnindice, 7)>>
          IF !EMPTY(ltstruct( m.lnindice, 8))
            \*<<SPACE(m.lnlong+9)>><<ltstruct( m.lnindice, 8)>>
          ENDIF && !EMPTY(ltstruct( m.lnindice, 8))
        ELSE && !EMPTY(ltstruct( m.lnindice, 7))
          \*  <<STR(m.lnindice,3)>> <<PADR(ltstruct( m.lnindice, 1), m.lnlong)>> : 
          \\<<ltstruct( m.lnindice, 8)>>
        ENDIF && !EMPTY(ltstruct( m.lnindice, 7))
      ENDIF && !EMPTY(ltstruct( m.lnindice, 7))
    NEXT && m.lnindice = 1 TO m.lnnbchamp
  ENDIF && m.llaccu
  \*
*!*  AFIELDS : une ligne est cre pour chaque champ de la table.
*!*            la premire ligne contient des informations sur la table (champs 10  18)
*!*  Col  info      type de donnes  
*!*  1  Field name  Character
*!*  2  Field type: 
*!*     C = Character     Y = Currency
*!*     D = Date          T = DateTime
*!*     B = Double        F = Float
*!*     G = General       I = Integer
*!*     L = Logical       M = Memo
*!*     N = Numeric       Q = Varbinary
*!*     V = Varchar and Varchar (Binary)
*!*     W = Blob  Character
*!*  3 Field width  Numeric
*!*  4 Decimal places  Numeric
*!*  5 Null values allowed  Logical
*!*  6 Code page translation not allowed  Logical
*!*  7 Field validation expression Character
*!*  8 Field validation text Character
*!*  9 Field default value Character
*!*  10 Table validation expression Character
*!*  11 Table validation text Character
*!*  12 Long table name Character
*!*  13 Insert trigger expression Character
*!*  14 Update trigger expression  Character
*!*  15 Delete trigger expression  Character
*!*  16 Table comment  Character 
*!*  17 NextValue for autoincrementing Numeric
*!*  18 Step for autoincrementing  Numeric
  lnaccu = ATAGINFO( lttag)
  IF m.lnaccu = 0
    \*   Pas d'index
  ELSE && m.lnaccu = 0
    \*   Index:
*!*  ATAGINFO() : informations sur les index de la table
*!*  Col  Description  Data type  
*!*  1 Index tag name in a .cdx file or index file name for .idx files  Character
*!*  2 Index tag type  Character
*!*  3 Index key expression  Character
*!*  4 Filter expression  Character
*!*  5 Index order as created (Ascending or Descending) Character
*!*  6 Collate sequence  Character
    STORE 0 TO lnlong, lnlong2 && pour bien aligner les colonnes ...
    FOR  m.lnindice = 1 TO m.lnaccu
      lnlong = MAX(m.lnlong, LEN(lttag(m.lnindice, 1)))
      lnlong2 = MAX(m.lnlong2, LEN(lttag(m.lnindice, 3)))
    NEXT && m.lnindice = 1 TO m.lnaccu
    lnlong = m.lnlong + 1
    lnlong2 = m.lnlong2 + 1
    FOR  m.lnindice = 1 TO m.lnaccu
      \*     tag: <<PADR(lttag(m.lnindice, 1), m.lnlong)>>  <<PADR(lttag(m.lnindice, 2), 10)>>  
      \\clef: <<PADR(lttag(m.lnindice, 3), m.lnlong2)>>
      \\  <<lttag(m.lnindice, 5)>>  <<lttag(m.lnindice, 6)>>
      IF !EMPTY(lttag(m.lnindice, 4))
        \*      filtre: <<lttag(m.lnindice, 4)>>
      ENDIF && !EMPTY(lttag(m.lnindice, 4))
    NEXT && m.lnindice = 1 TO m.lnaccu
    \*
  ENDIF && m.lnaccu = 0
  * on essaie de mettre la liste des vues sur cette table
  \*  Liste des vues sur cette table uniquement
  lnnombre = 0
  FOR  m.lnindice = 1 TO m.nTotal_Views
    * on a une vue mais concerne-t-elle la table courante ?
    lcaccu = LOWER(ALLTRIM(DBGetProp(aAll_Views[m.lnindice], 'View', 'Table')))
    * VFP met le nom de la BDD devant chaque nom de table ....
    lnaccu = AT( "!", m.lcaccu)
    IF m.lnaccu > 0
      lcaccu = SUBSTR(m.lcaccu, m.lnaccu+1)
    ENDIF && m.lnaccu > 0
    lnaccu = AT(",", m.lcaccu)
    IF m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
      * une seule table
      lnnombre = m.lnnombre + 1
      lcnomvue = ALLTRIM(LOWER(aAll_Views[m.lnindice]))
      lcrequete = g_ltsqlvue( m.lnindice)
      llupdatable = DBGetProp(m.lcnomvue, 'View', 'SendUpdates')
      \*    <<m.lcnomvue>>  (<<IIF(m.llupdatable, "lecture/criture", "lecture seule")>>)
      * on est oblig d'imprimer ligne  ligne pour mettre l'toile devant
      lnligne = ALINES( lttab, m.lcrequete)
      FOR m.lni = 1 TO m.lnligne
        \*      <<lttab(m.lni)>>
      NEXT && m.lni = 1 TO m.lnligne
      \*
    ENDIF && m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
  NEXT && m.lnindice = 1 TO m.nTotal_Views
  IF m.lnnombre = 0
    \*    aucune
  ENDIF && m.lnnombre = 0

  \*  Liste des vues sur plusieurs tables dont celle-ci
  lnnombre = 0
  FOR  m.lnindice = 1 TO m.nTotal_Views
    * on a une vue mais concerne-t-elle la table courante ?
    lcaccu = LOWER(ALLTRIM(DBGetProp(aAll_Views[m.lnindice], 'View', 'Table')))
    * VFP met le nom de la BDD devant chaque nom de table ....
    lnaccu = AT( "!", m.lcaccu)
    IF m.lnaccu > 0
      lcaccu = SUBSTR(m.lcaccu, m.lnaccu+1)
    ENDIF && m.lnaccu > 0
    lnaccu = AT(",", m.lcaccu)
    IF m.lnaccu > 0 AND AT(m.lcnomtable, m.lcaccu)>0 
      * plusieurs tables et la table courante est dans la liste
      lnnombre = m.lnnombre + 1
      lcnomvue = ALLTRIM(LOWER(aAll_Views[m.lnindice]))
      lcrequete = g_ltsqlvue( m.lnindice)
      llupdatable = DBGetProp(m.lcnomvue, 'View', 'SendUpdates')
      \*    <<m.lcnomvue>>  (<<IIF(m.llupdatable, "lecture/criture", "lecture seule")>>)
      lnligne = ALINES( lttab, m.lcrequete)
      FOR  m.lni = 1 TO m.lnligne
        \*      <<lttab(m.lni)>>
      NEXT && m.lni = 1 TO m.lnligne
      \*
    ENDIF && m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
  NEXT && m.lnindice = 1 TO m.nTotal_Views
  IF m.lnnombre = 0
    \*    aucune
  ENDIF && m.lnnombre = 0

*!*    FOR  m.lnindice = 1 TO m.nTotal_Views
*!*      IF RIGHT(ALLTRIM(UPPER(aAll_Views[m.lnindice])), 3) == "_RW"
*!*        * on a une vue _RW, mais concerne-t-elle la table courante ?
*!*        lcaccu = LOWER(ALLTRIM(DBGetProp(aAll_Views[m.lnindice], 'View', 'Table')))
*!*        * VFP met le nom de la BDD devant chaque nom de table ....
*!*        lnaccu = AT( "!", m.lcaccu)
*!*        IF m.lnaccu > 0
*!*          lcaccu = SUBSTR(m.lcaccu, m.lnaccu+1)
*!*        ENDIF && m.lnaccu > 0
*!*  *!*          \* <<aAll_Views[m.lnindice]>> table: <<m.lcaccu>> = <<m.lcnomtable>>
*!*        lnaccu = AT(",", m.lcaccu)
*!*        IF m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
*!*          * une seule table et c'est la bonne
*!*          llvuerw = .T.
*!*          lcvuerw = ALLTRIM(LOWER(aAll_Views[m.lnindice]))
*!*          lcrequete = g_ltsqlvue( m.lnindice)
*!*          * il faut chercher le ou les champs clefs parce qu'on va les utiliser 
*!*          * plus loin dans la recherche d'une nouvelle clef dans ajout()
*!*          * on ouvre la table, puis la vue _RW pour en rcuprer la liste des champs et
*!*          * le ou les champs qui interviennent dans la clef pour pouvoir gnrer
*!*          * le programme de calcul de la nouvelle valeur de cette clef  partir de la
*!*          * table clef.dbf
*!*          * ATTENTION on est dans le programme de gnration !
*!*          SELECT 0
*!*          USE (m.lcnomtable) SHARED AGAIN NOUPDATE
*!*          SELECT 0
*!*          USE (m.lcvuerw) NODATA
*!*          IF !EMPTY(ALIAS())
*!*            lnnbchamps = AFIELDS(ltchamp)
*!*          ENDIF && !EMPTY(ALIAS())
*!*          USE IN (SELECT(m.lcvuerw))
*!*          USE IN (SELECT(m.lcnomtable))
*!*          EXIT
*!*        ENDIF && m.lnaccu = 0 AND m.lcaccu = m.lcnomtable
*!*      ENDIF && RIGHT(ALLTRIM(UPPER(aAll_Views[m.lnindice])), 3) == "_RW"
*!*    NEXT && m.lnindice = 1 TO m.nTotal_Views
  
  
  USE
  RETURN
ENDPROC && informations_table

FUNCTION lettre_vers_type(pccharac AS Character) AS String && 10/02/2009
  LOCAL lcresult
  
  DO CASE 
  CASE m.pccharac == "C"
    lcresult = "Character"
  CASE m.pccharac == "D"
    lcresult = "Date"
  CASE m.pccharac == "T"
    lcresult = "DateTime"
  CASE m.pccharac == "N"
    lcresult = "Numeric"
  CASE m.pccharac == "I"
    lcresult = "Integer"
  CASE m.pccharac == "L"
    lcresult = "Logical"
  CASE m.pccharac == "M"
    lcresult = "Memo"
  CASE m.pccharac == "G"
    lcresult = "General"
  CASE m.pccharac == "Y"
    lcresult = "Currency"
  CASE m.pccharac == "B"
    lcresult = "Double"
  CASE m.pccharac == "F"
    lcresult = "Float"
  CASE m.pccharac == "Q"
    lcresult = "Varbinary"
  CASE m.pccharac == "V"
    lcresult = "Varchar"
  CASE m.pccharac == "W"
    lcresult = "blob"
  OTHERWISE
    lcresult = "erreur"
  ENDCASE
  RETURN m.lcresult
ENDFUNC && lettre_vers_type
